/**
	$Id:JspContext.java 4187 2008-03-20 11:15:25Z nicklas $

	Copyright (C) Authors contributing to this file.

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.extensions;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import javax.servlet.jsp.JspWriter;
import javax.servlet.jsp.PageContext;

import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.SessionControl;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.util.extensions.ActionFactory;
import net.sf.basedb.util.extensions.ClientContext;
import net.sf.basedb.util.extensions.Extension;
import net.sf.basedb.util.extensions.InvokationContext;
import net.sf.basedb.util.extensions.RendererFactory;

/**
	Context object for the web application. Instances of this class
	are always used for passing client context information to web extensions.
	This means that it is always safe to cast the client context to this
	class in the {@link ActionFactory#prepareContext(InvokationContext)} 
	and {@link ActionFactory#getActions(InvokationContext)} methods
	and other methods that has a context parameter.
	
	<pre class="code">
&#64;Override
public boolean prepareContext(InvokationContext context) 
{
	JspContext jspContext = (JspContext)context.getClientContext();
	jspContext.addStylesheet("/base/include/styles/mystylesheet.css");
	// Do other things
	return true;
}
</pre>

	<p>
	This context exposes the underlying JSP {@link PageContext},
	which can be used to get information about the current request
	among other things. Render factories usually needs access to
	the response output stream, {@link #getOut()}, so that renderers 
	can write the generated HTML to the response.

	@author nicklas
	@version 2.7
	@base.modified $Date:2008-03-20 12:15:25 +0100 (Thu, 20 Mar 2008) $
*/
public class JspContext
	extends ClientContext
{
	
	/**
		The key that is used when storing a JspContext object in
		a request/page/session/etc. attribute.
		@since 2.17
	*/
	public static final String ATTRIBUTE_KEY = "net.sf.basedb.clients.web.extensions.JspContext";
	
	/**
		The key that is used when storing an Extension object in
		a request/page/session/etc. attribute.
		@since 3.3
	*/
	public static final String EXTENSION_KEY = "net.sf.basedb.util.extensions.Extension";
	
	private final PageContext pageContext;
	private final GuiContext guiContext;
	
	private boolean needResourcesPerExtension;
	private Map<String, Set<String>> scripts;
	private Map<String, Set<String>> stylesheets;
	
	JspContext(SessionControl sc, DbControl dc, 
		PageContext pageContext, GuiContext guiContext, Object item)
	{
		super(sc, dc, item);
		this.pageContext = pageContext;
		this.guiContext = guiContext;
	}

	/**
		Set a flag indicating that resources (eg. scripts and stylesheets)
		should be tracked per extension. If this is set, it is possible to
		use {@link #getScripts(String)} and {@link #getStylesheets(String)}
		to get the resources that was added by the given extension.
		@since 3.19.3
	*/
	public void setNeedResourcesPerExtension(boolean nrpe)
	{
		this.needResourcesPerExtension = nrpe;
	}
	
	/**
		Get the JSP Page context object for the current request.
	*/
	public PageContext getPageContext()
	{
		return pageContext;
	}

	/**
		Get a writer object that can be used to return generated text
		to the response that is sent back to the browser.
		<p>
		
		Note! The writer is only intended to be used by render
		objects when rendering the extensions. Using the writer
		from the {@link ActionFactory#prepareContext(InvokationContext)} or 
		{@link RendererFactory#prepareContext(InvokationContext)}
		method may produce unpredictable results.
		
		@return A {@link JspWriter} object
	*/
	public JspWriter getOut()
	{
		return pageContext.getOut();
	}
	
	/**
		Get the current {@link GuiContext}.
		@return A GuiContext object or null if it is not applicable
			to the extension point
		@since 2.12
	*/
	public GuiContext getGuiContext()
	{
		return guiContext;
	}
	
	/**
		Get the URL to the web application root as a string. The URL
		only includes the path information, not the server name or protocol.
	*/
	public String getRoot()
	{
		return pageContext.getServletContext().getContextPath();
	}
	
	/**
		Get the URL to the home directory for a given extension. The URL
		only includes the path information, not the server name or protocol.
		@param extension The extension
	*/
	public String getHome(Extension<?> extension)
	{
		return ExtensionsControl.getHomeUrl(extension.getId());
	}
	
	/**
		Add an absolute reference to a script that should be included on the
		generated page with a <code>&lt;script&gt;</code> tag. The 
		path must be an absolute path. 

		<p>
		Note 1! This method must be called from the {@link 
		ActionFactory#prepareContext(InvokationContext)} or 
		{@link RendererFactory#prepareContext(InvokationContext)}
		methods. Changes to the context after that will not be reflected
		in the output.
		
		<p>
		Note 2! Not all extension points supports custom scripts. Check
		the documentation for the extension point you are extending.
		
		@param absolutePath The absolute path
	*/
	public void addScript(String absolutePath)
	{
		if (scripts == null) scripts = new HashMap<>();
		getSet(scripts, null).add(absolutePath);
		if (needResourcesPerExtension)
		{
			getSet(scripts, getCurrentExtension().getId()).add(absolutePath);
		}
	}
	
	/**
		Add several scripts in one go. If the collection is null or empty
		nothing is done.
		@since 3.19.4
	*/
	public void addScripts(Collection<String> scriptsToAdd)
	{
		if (scriptsToAdd == null || scriptsToAdd.isEmpty()) return;
		if (scripts == null) scripts = new HashMap<>();
		getSet(scripts, null).addAll(scriptsToAdd);
		if (needResourcesPerExtension)
		{
			getSet(scripts, getCurrentExtension().getId()).addAll(scriptsToAdd);
		}
	}
	
	/**
		Get the Set that is stored under the given key. If
		no Set exists it is created and stored in the map.
	*/
	private Set<String> getSet(Map<String, Set<String>> map, String key)
	{
		Set<String> set = map.get(key);
		if (set == null)
		{
			set = new HashSet<>();
			map.put(key, set);
		}
		return set;
	}
	
	/**
		Add an absolute reference to a CSS stylesheet that should be included on the
		generated page with a <code>&lt;link&gt;</code> tag. The 
		path must be the absolute path. 

		<p>
		Note 1! This method must be called from the {@link 
		ActionFactory#prepareContext(InvokationContext)} or 
		{@link RendererFactory#prepareContext(InvokationContext)}
		methods. Changes to the context after that will not be reflected
		in the output.
		
		<p>
		Note 2! Not all extension points supports custom stylesheets. Check
		the documentation for the extension point you are extending.

		@param absolutePath The absolute path
	*/
	public void addStylesheet(String absolutePath)
	{
		if (stylesheets == null) stylesheets = new HashMap<>();
		getSet(stylesheets, null).add(absolutePath);
		if (needResourcesPerExtension)
		{
			getSet(stylesheets, getCurrentExtension().getId()).add(absolutePath);
		}
	}
	
	/**
		Add several stylesheets in one go. If the collection is null or empty
		nothing is done.
		@since 3.19.4
	*/
	public void addStylesheets(Collection<String> stylesheetsToAdd)
	{
		if (stylesheetsToAdd == null || stylesheetsToAdd.isEmpty()) return;
		if (stylesheets == null) stylesheets = new HashMap<>();
		getSet(stylesheets, null).addAll(stylesheetsToAdd);
		if (needResourcesPerExtension)
		{
			getSet(stylesheets, getCurrentExtension().getId()).addAll(stylesheetsToAdd);
		}
	}

	/**
		Get all scripts that has been added to this context.
		@return A collection of scripts
	*/
	public Collection<String> getScripts()
	{
		return scripts == null ? null : scripts.get(null);
	}
	
	/**
		Get all scripts that has been added to this context
		by the given extension
		@return A collection of scripts
		@since 3.19.3
	*/
	public Collection<String> getScripts(String xtId)
	{
		return scripts == null ? null : scripts.get(xtId);
	}
	
	/**
		Get all stylesheets that has been added to this context.
		@return A collection of stylesheets
	*/
	public Collection<String> getStylesheets()
	{
		return stylesheets == null ? null : stylesheets.get(null);
	}
	
	/**
		Get all stylesheets that has been added to this context
		by the given extension.
		@return A collection of stylesheets
		@since 3.19.3
	*/
	public Collection<String> getStylesheets(String xtId)
	{
		return stylesheets == null ? null : stylesheets.get(xtId);
	}
	
}
