package net.sf.basedb.clients.web.extensions.list;

import java.util.Collections;
import java.util.Set;
import java.util.TreeSet;

import net.sf.basedb.core.Annotatable;
import net.sf.basedb.core.AnnotationType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemList;
import net.sf.basedb.core.ItemSubtype;
import net.sf.basedb.core.Metadata;
import net.sf.basedb.core.RawDataTypes;
import net.sf.basedb.core.StringUtil;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restriction;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.Metadata.PropertyPath;
import net.sf.basedb.core.RawDataType;
import net.sf.basedb.core.SyncFilter.SourceItemTransform;
import net.sf.basedb.util.NameableComparator;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.extensions.ExtensionsInvoker;
import net.sf.basedb.util.filter.Filter;
import net.sf.basedb.util.listable.ListableUtil;
import net.sf.basedb.util.listable.RestrictionTransformer;
import net.sf.basedb.util.listable.SourceItemTransformer;
import net.sf.basedb.util.listable.SourceItemTransformerFactory;

/**
	A column action implementation that finds a parent or child item of a given 
	subtype and	then get an annotation or other property value from it.
	
	@author nicklas
	@since 3.16, 3.17
*/
public abstract class RelatedItemColumn
	extends AbstractListColumnBean<Annotatable, Object>
{
	
	/**
		For now, we assume that 'expr' is one of:
		/DIRECTION/ITEMTYPE/subtype-id/#annotationtype-id
		/DIRECTION/ITEMTYPE/subtype-id/!extension-id#action-id
		/DIRECTION/ITEMTYPE/subtype-id/property
		
		DIRECTION=CHILD|PARENT (PARENT is assumed if not specified)
		'subtype-id' can also be the rawdatatype-id if the ITEMTYPE=Item.RAWBIOASSAY
	*/
	@SuppressWarnings({ "unchecked", "rawtypes" })
	public static RelatedItemColumn create(DbControl dc, int index, Specification spec, RelatedItemHelper helper)
	{
		RelatedItemColumn col = null;
		try
		{
			if (spec.property.startsWith("/"))
			{
				// This is a multi-hop path: child --> parent --> child
				if (spec.direction != SourceItemTransform.CHILD_TO_PARENT)
				{
					throw new RuntimeException("First multi-hop must be to /PARENT: " + spec.expression);
				}
				Specification nextSpec = Specification.parse(spec.targetType, spec.property);
				if (nextSpec.property.startsWith("/"))
				{
					throw new RuntimeException("More than two PARENT/CHILD hops are not allowed: " + spec.expression);
				}
				if (nextSpec.direction != SourceItemTransform.PARENT_TO_CHILD)
				{
					throw new RuntimeException("Last multi-hop must be to /CHILD: " + spec.expression);
				}
				RelatedItemColumn nextHop = create(dc, index, nextSpec, helper);
				col = new RelatedItemMultiHopColumn(dc, index, spec, helper, nextHop); 
			}
			else if (spec.property.startsWith("#"))
			{
				AnnotationType at = AnnotationType.getById(dc, spec.annotationId);
				col = new RelatedItemAnnotationColumn(dc, index, spec, helper, at);
			}
			else if (spec.property.startsWith("!"))
			{
				ListColumnAction <Annotatable, ?> action = null;
				// Get the invoker that was saved by the RelatedItemColumnActionFactory.prepareContext() method
				ExtensionsInvoker<ListColumnAction<Annotatable, ?>> invoker = helper.jspContext.getAttribute("RelatedItemInvoker."+spec.expression);
				
				if (invoker != null)
				{
					invoker.getClientContext().setCurrentItem(spec);
					for (ListColumnAction<Annotatable, ?> lca : invoker)
					{
						if (spec.actionId.equals(lca.getId()) || spec.actionId.equals(lca.getFilterProperty()))
						{
							action = lca;
							break;
						}
					}
				}
				if (action == null) throw new RuntimeException("Extension not found: "+Values.getString(spec.extensionId)+"#"+spec.actionId);
				col = new RelatedItemExtensionColumn(dc, index, spec, helper, action);
			}
			else if (spec.property.equals("§itemLists"))
			{
				Filter<? super ItemList> itemListFilter = null;
				if (spec.targetType == Item.RAWBIOASSAY)
				{
					itemListFilter = new RawDataTypeFilter(spec.getRawDataType());
				}
				else
				{
					itemListFilter = new SubtypeFilter(spec.getItemSubtype(dc));
				}
				col = new RelatedItemListColumn(dc, index, spec, itemListFilter, helper);
			}
			else
			{
				PropertyPath property = null;
				if (!".".equals(spec.property)) // special case '.' is equal to the "related" item
				{
					Metadata m = Metadata.getInstance(spec.targetType.getItemClass());
					property = m.getPropertyPath(spec.property, false);
				}
				col = new RelatedItemPropertyColumn(dc, index, spec, helper, property);
			}
		}
		catch (RuntimeException ex)
		{
			ex.printStackTrace(System.out);
			col = new RelatedItemErrorColumn(index, spec, helper, ex.getMessage());

		}
		return col;
	}

	private final int index;
	private final Specification spec;
	private final RelatedItemHelper helper;
	private final SourceItemTransformer transformer;
	
	RelatedItemColumn(DbControl dc, int index, Specification spec, RelatedItemHelper helper)
	{
		this.index = index;
		this.helper = helper;
		this.spec = spec;
		if (dc != null)
		{
			boolean includeChildrenThatPushToParents =  !spec.multiHop && 
					(spec.getTargetType() == Item.EXTRACT || spec.getTargetType() == Item.SAMPLE) && 
					spec.getItemSubtype(dc).getPushAnnotations();
			SourceItemTransformerFactory transformerFactory = ListableUtil.getTransformerFactory(spec.targetType, includeChildrenThatPushToParents);
			SourceItemTransformer tmp = transformerFactory.create(spec.sourceType, spec.direction);
			Restriction targetRestriction = spec.createTargetTypeRestriction(dc);
			if (targetRestriction != null) tmp = new RestrictionTransformer(tmp, targetRestriction);
			this.transformer = tmp;
		}
		else
		{
			this.transformer = null;
		}
		setId(spec.expression);
		setProperty(spec.expression);
		setExportable(true);
		setFilterable(true);
		setClazz("relateditemcol" + 
				(spec.direction == SourceItemTransform.CHILD_TO_PARENT ? " parentitem" : " childitem")+
				(spec.multiHop ? " multihop" : "")
				);
	}

	/**
		Is this column lazy-loaded? The default implementation 
		use the requested laza flag from the helper instance.
		@since 3.19.4
	*/
	protected boolean isLazy()
	{
		return helper.lazy;
	}
	
	/**
		Get the transformer implementation that is used to turn source item into target items.
		@since 3.19.4
	*/
	protected SourceItemTransformer getTransformer()
	{
		return transformer;
	}
	
	/**
		Get the specification for this column.
		@since 3.19.5
	*/
	protected Specification getSpecification()
	{
		return spec;
	}
	
	/**
		We finalize this implementation to make sure that the helper 
		implementation always get a chance to re-cycle transactions.
		Subclasses should implement {@link #getValue(RelatedItemHelper, Annotatable, SourceItemTransformer)}
		to return the requested value from the item and use the
		DbControl from {@link RelatedItemHelper#dc}.
	*/
	@Override
	public final Object getValue(DbControl dc, Annotatable item) 
	{
		if (isLazy()) return getProxyTag(item);
		item = helper.recycle(dc, item);
		return getValue(helper, item, null);
	}

	/**
		Generate a proxy tag that is returned instead of actual data. A
		JavaScript in 'lazy-parent-items.js' will scan the table and
		inject the actual data in batches.
	*/
	private String getProxyTag(Annotatable item)
	{
		StringBuilder sb = new StringBuilder();
		sb.append("<div");
		sb.append(" class=\"lazy-item not-loaded\"");
		sb.append(" id=\"lazy-item-").append(index).append("-").append(item.getId()).append("\"");
		sb.append(" data-item-id=\"").append(item.getId()).append("\"");
		sb.append(" data-index=\"").append(index).append("\"");
		sb.append("></div>");
		return sb.toString();
	}
	
	/**
		We finalize this implementation to make sure that the helper 
		implementation always get a chance to re-cycle transactions.
		Subclasses should override {@link #getExportValue(RelatedItemHelper, Annotatable, SourceItemTransformer)}
		to return the requested value from the item and use the
		DbControl from {@link RelatedItemHelper#dc}.
	*/
	@Override
	public final Object getExportValue(DbControl dc, Annotatable item) 
	{
		item = helper.recycle(dc, item);
		return getExportValue(helper, item, null);
	}

	
	/**
		Alternate implementation for loading data. The item is normally an item of the
		expected source type, but if a preTransform is present, the item may be of 
		different type and must be processed with the {@link SourceItemTransformer#transform(net.sf.basedb.util.listable.TransformContext, Set)}
		method first. In most cases, implementation should simply call {@link #getRelatedItems(Annotatable, SourceItemTransformer)}
		to load the final target items.
	*/
	protected abstract Object getValue(RelatedItemHelper helper, Annotatable item, SourceItemTransformer preTransform);
	
	/**
		Alternate implementation for loading data. The default implementation
		simply call the {@link #getValue(RelatedItemHelper, Annotatable, SourceItemTransformer)} method.
		@since 3.19.4
	*/
	protected Object getExportValue(RelatedItemHelper helper, Annotatable item, SourceItemTransformer preTransform)
	{
		return getValue(helper, item, preTransform);
	}
	
	/**
		Get all parent or child items to the requested item that matches the specified filter.
	*/
	@SuppressWarnings({ "unchecked", "rawtypes" })
	protected Set<Annotatable> getRelatedItems(Annotatable item, SourceItemTransformer preTransform)
	{
		Set<Integer> sourceIds = Collections.singleton(item.getId());
		helper.transformContext.resetCollected();
		if (preTransform != null)
		{
			if (!spec.allowDoublingBack) helper.transformContext.setCollecting();
			sourceIds = preTransform.transform(helper.transformContext, sourceIds);
			if (!spec.allowDoublingBack) helper.transformContext.setAvoiding();
		}
		Set<Integer> relatedIds = transformer.transform(helper.transformContext, sourceIds);
		
		Set<Annotatable> relatedItems = new TreeSet<Annotatable>(new NameableComparator(false));
		for (Integer id : relatedIds)
		{
			Annotatable a = spec.targetType.getById(helper.dc, id);
			relatedItems.add(a);
		}
		return relatedItems;
	}
	
	/**
		Represents a specification for loading a related parent or child item.
		The general format is: /DIRECTION/ITEMTYPE/subtype-id/property
		
		where 
		* DIRECTION: PARENT or CHILD
		* ITEMTYPE: Item.name() value for the related item
		* subtype-id: Id of an {@link ItemSubtype} or a {@link RawDataType}
		* property: a specification of what value to retrieve from the related item.
		  It can have different formats:
		  - #annotationtype-id: Id of an annotation type
		  - !extension-id#action-id: Id of an extension and {@link ListColumnAction} to use for loading the value
		  Otherwise it is interpreted as a property that is defined on the targeted instance.
		
		@since 3.19.4
	*/
	public static class Specification
	{
		
		public static Specification parse(Item sourceType, String expression)
		{
			Specification spec = new Specification();
			spec.sourceType = sourceType;
			spec.expression = expression;
			spec.directionRaw = "PARENT";
			spec.direction = SourceItemTransform.CHILD_TO_PARENT;
			
			String[] tmp = expression.split("/", 5);
			int baseIndex = 1;
			if (tmp.length == 5)
			{
				baseIndex = 2;
				spec.directionRaw = tmp[1];
				if (spec.directionRaw.endsWith("+")) spec.allowDoublingBack = true;
				if (spec.directionRaw.startsWith("CHILD")) spec.direction = SourceItemTransform.PARENT_TO_CHILD;
			}
			spec.targetType = Item.valueOf(tmp[baseIndex]);
			spec.subtype = tmp[baseIndex+1];
			spec.property = tmp[baseIndex+2];
			
			if (spec.property.startsWith("/"))
			{
				spec.multiHop = true;
			}
			else if (spec.property.startsWith("#"))
			{
				spec.annotationId = Values.getInt(spec.property.substring(1));
			}
			else if (spec.property.startsWith("!"))
			{
				// The normal property format is: !extension-id#action-id
				// but it may be a filter specification: !x.action-id
				// or (but it should never happen): !action-id
				if (spec.property.startsWith("!x."))
				{
					spec.actionId = spec.property.substring(3);
				}
				else
				{
					int splitAt = spec.property.indexOf('#');
					if (splitAt > 0)
					{
						spec.extensionId = Values.getStringOrNull(spec.property.substring(1, splitAt));
						spec.actionId = spec.property.substring(splitAt+1);
					}
					else
					{
						spec.actionId = spec.property.substring(1);
					}
				}
			}
			
			return spec;
		}
		
		String expression;
		String directionRaw;
		SourceItemTransform direction;
		boolean allowDoublingBack;
		Item sourceType;
		Item targetType;
		String subtype;
		String subtypeName;
		String property;
		int annotationId;
		String extensionId;
		String actionId;
		boolean multiHop;
		
		/**
			Get the full expression that defines the targeted item and property.
		*/
		public String getExpression() 
		{
			return expression;
		}

		/**
			Get the raw value of the DIRECTION.
		*/
		public String getDirectionRaw() 
		{
			return directionRaw;
		}

		/**
			Get the parsed value of the DIRECTION.
		*/
		public SourceItemTransform getDirection() 
		{
			return direction;
		}

		/**
			Enabled on a multi-hop path if the path down to the related child from the
			parent item is allowed to traverse the same items as the path up from
			the source child item to the parent.
		*/
		public boolean isDoublingBackAllowed()
		{
			return allowDoublingBack;
		}
		
		/**
			Get the source item type. This is not part of the expression, but
			is taken from the current list.
		*/
		public Item getSourceType() 
		{
			return sourceType;
		}

		/**
			Get the target ITEMTYPE from the expression.
		*/
		public Item getTargetType() 
		{
			return targetType;
		}

		/**
			Get the target subtype from the expression.
		*/
		public String getSubtype() 
		{
			return subtype;
		}

		/**
			Get the name of the targeted subtype. This may not always
			be available.
		*/
		public String getSubtypeName() 
		{
			return subtypeName;
		}

		/**
			Get the property from the expression.
		*/
		public String getProperty() 
		{
			return property;
		}

		/**
			Get the annotation id from the expression. Only available if the 
			property represents an annotation (it starts with '#').
		*/
		public int getAnnotationId() 
		{
			return annotationId;
		}

		/**
			Get the extension id from the expression. Only available if the 
			property represents an extension (it starts with '!').
		*/
		public String getExtensionId() 
		{
			return extensionId;
		}

		/**
			Get the action id from the expression. Only available if the 
			property represents an extension (it starts with '!').
		*/
		public String getActionId() 
		{
			return actionId;
		}
		
		/**
			If the property starts with a '/' it is a multi-hop specification.
			A multi-hop specification must be from child->parent->child and must
			not have more than two hops.
		*/
		public boolean isMultiHop()
		{
			return multiHop;
		}
		
		/**
			Generates a title using target item type and subtype as prefix.
		*/
		public String generateTitle(String title)
		{
			String result = "<span class=\"";
			result += direction == SourceItemTransform.CHILD_TO_PARENT ? "parentitem" : "childitem";
			if (multiHop) result += " multihop";
			if (allowDoublingBack) result += " doublingback";
			result += "\">";
			result += StringUtil.coalesce(subtypeName, targetType.name());
			result += "</span>";
			result += withPrefix(".", title);
			return result;
		}
		
		/**
			Generates a tooltip using target item type and subtype as prefix.
		*/
		public String generateTooltip(String tooltip)
		{
			return (direction == SourceItemTransform.CHILD_TO_PARENT ? "Parent › " : "Child › ") + 
					StringUtil.coalesce(subtypeName, targetType.name()) + withPrefix(" › ", tooltip);
		}
		
		/**
			Return prefix+string if string is not null or the empty string otherwise.
		*/
		private String withPrefix(String prefix, String s)
		{
			return s == null ? "" : prefix+s;
		}

		public ItemSubtype getItemSubtype(DbControl dc)
		{
			ItemSubtype st = ItemSubtype.getById(dc, Values.getInt(subtype));
			subtypeName = st.getName();
			return st;
		}
		
		public RawDataType getRawDataType()
		{
			RawDataType rdt = RawDataTypes.getSafeRawDataType(subtype);
			subtypeName = rdt.getName();
			return rdt;
		}
		
		public Restriction createTargetTypeRestriction(DbControl dc)
		{
			Restriction r = null;
			if (targetType == Item.RAWBIOASSAY)
			{
				RawDataType rdt = getRawDataType();
				r = Restrictions.eq(Hql.property("rawDataType"), Expressions.string(rdt.getId()));
			}
			else
			{
				ItemSubtype subtype = getItemSubtype(dc);
				r = Restrictions.eq(Hql.property("itemSubtype"), Hql.entity(subtype));
			}
			return r;
		}

	}
	
}