/*
	$Id: HibernateUtil.java 8134 2023-03-31 09:32:24Z nicklas $

	Copyright (C) 2005 Samuel Andersson, Johan Enell, Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson
	Copyright (C) 2007 Johan Enell, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.AnnotatableData;
import net.sf.basedb.core.data.AnnotationData;
import net.sf.basedb.core.data.BasicData;
import net.sf.basedb.core.data.MessageData;
import net.sf.basedb.core.data.RemovableData;
import net.sf.basedb.core.data.SessionData;
import net.sf.basedb.core.data.OwnableData;
import net.sf.basedb.core.data.ShareableData;
import net.sf.basedb.core.data.ExtendableData;
import net.sf.basedb.core.data.NewsData;
import net.sf.basedb.core.data.UserData;
import net.sf.basedb.core.data.UserDeviceData;
import net.sf.basedb.core.dbengine.DbEngine;
import net.sf.basedb.core.dbengine.EngineFactory;
import net.sf.basedb.core.dbengine.TableInfo;
import net.sf.basedb.core.dbengine.TableInfo.ColumnInfo;
import net.sf.basedb.core.dbengine.TableInfo.ForeignKeyInfo;
import net.sf.basedb.core.dbengine.TableInfo.IndexInfo;
import net.sf.basedb.core.hibernate.DbIndexWork;
import net.sf.basedb.core.hibernate.EntityQueryWrapper;
import net.sf.basedb.core.hibernate.ExecuteUpdateWork;
import net.sf.basedb.core.hibernate.SchemaGenerator;
import net.sf.basedb.core.hibernate.SessionWrapper;
import net.sf.basedb.core.hibernate.TypeWrapper;
import net.sf.basedb.core.hibernate.MultiUpdateWork;
import net.sf.basedb.core.hibernate.NativeQueryWrapper;
import net.sf.basedb.core.hibernate.NativeSQLFunction;
import net.sf.basedb.core.hibernate.TableExistsWork;
import net.sf.basedb.core.query.QueryType;
import net.sf.basedb.util.ClassUtil;
import net.sf.basedb.util.FileUtil;
import net.sf.basedb.util.RegexpFileFilter;
import net.sf.basedb.util.XmlUtil2;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Iterator;
import java.util.Date;
import java.util.Map;
import java.util.Properties;
import java.util.ServiceLoader;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;

import javax.persistence.PersistenceException;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.Statement;
import java.sql.ResultSet;
import java.sql.DatabaseMetaData;
import java.sql.Driver;
import java.sql.SQLException;
import java.io.IOException;
import java.net.URL;
import java.nio.charset.StandardCharsets;

import org.jdom2.Document;
import org.jdom2.Element;
import org.jdom2.Namespace;
import org.jdom2.output.XMLOutputter;
import org.hibernate.Hibernate;
import org.hibernate.Interceptor;
import org.hibernate.LockOptions;
import org.hibernate.SessionBuilder;
import org.hibernate.SessionFactory;
import org.hibernate.Session;
import org.hibernate.StatelessSession;
import org.hibernate.Transaction;
import org.hibernate.HibernateException;
import org.hibernate.StaleStateException;
import org.hibernate.FlushMode;
import org.hibernate.query.Query;
import org.hibernate.query.NativeQuery;
import org.hibernate.Filter;
import org.hibernate.ScrollableResults;
import org.hibernate.ScrollMode;
import org.hibernate.exception.ConstraintViolationException;
import org.hibernate.internal.SessionImpl;
import org.hibernate.jdbc.ReturningWork;
import org.hibernate.jdbc.Work;
import org.hibernate.mapping.PersistentClass;
import org.hibernate.mapping.Table;
import org.hibernate.mapping.Index;
import org.hibernate.mapping.Column;
import org.hibernate.mapping.PrimaryKey;
import org.hibernate.boot.MetadataSources;
import org.hibernate.boot.SessionFactoryBuilder;
import org.hibernate.boot.archive.internal.ByteArrayInputStreamAccess;
import org.hibernate.boot.model.relational.Database;
import org.hibernate.boot.model.relational.SqlStringGenerationContext;
import org.hibernate.boot.model.relational.internal.SqlStringGenerationContextImpl;
import org.hibernate.boot.registry.StandardServiceRegistryBuilder;
import org.hibernate.cfg.Configuration;
import org.hibernate.metadata.ClassMetadata;
import org.hibernate.dialect.Dialect;
import org.hibernate.dialect.function.SQLFunction;
import org.hibernate.dialect.function.SQLFunctionTemplate;
import org.hibernate.engine.spi.Mapping;

/**
	This class collects most of the functionality we need for using Hibernate.

	@author Nicklas, Samuel
	@version 2.0
	@base.modified $Date: 2023-03-31 11:32:24 +0200 (Fri, 31 Mar 2023) $
*/
public class HibernateUtil
{

	/**
		Log core events.
	*/
	private static final org.slf4j.Logger log =
		org.slf4j.LoggerFactory.getLogger("net.sf.basedb.core");

	/**
		Log all batcher SQL statements.
	*/
	private static final org.slf4j.Logger logSql =
		org.slf4j.LoggerFactory.getLogger("net.sf.basedb.core.batcher.sql");


	/**
		The single SessionFactory.
	*/
	private static SessionFactory sf = null;

	/**
		The Hibernate configuration, including all mappings.
	*/
	private static MetadataSources cfg = null;
	
	/**
		The mappings related to a configuration.
	*/
	private static org.hibernate.boot.Metadata metadata = null;
	
	/**
		The Hibernate database dialect in use.
	*/
	private static Dialect dialect = null;
	
	/**
		Our extra database specific sql generation that is not possible by 
		the Hibernate dialect.
	*/
	private static DbEngine dbEngine = null;

	private static boolean isInitialised = false;
	
	/**
		Number of rows to load at a time when iterating over
		a result set.
		@since 3.9
	*/
	private static int jdbcFetchSize = 1000;

	/**
		First step of initialising this class. This is done at startup time by the
		{@link Application#start()} method. In this step we load configuration
		settings from the 'base.config' and 'hibernate.cfg.xml' files and
		create the Dialect and DbEngine objects.
		
		Initialising means that we
		load the configuration from the properties and the xml file,
		read all mapping files, generate additional mappings for the
		{@link ExtendableData} items and raw data, and generate filters
		used by {@link Query} implementation.
	*/
	static synchronized void init1()
		throws BaseException
	{
		// Return if we have already been initialised
		if (isInitialised) return;

		try
		{
			StandardServiceRegistryBuilder registry = new StandardServiceRegistryBuilder();
			
			// read base.config settings related to database connection and remap them to Hibernate setting
			Properties baseConfig = getConnectionProperties();
			registry.applySettings(baseConfig);
			
			// read hibernate.cfg.xml
			registry.configure();

			// Store configuration source files
			cfg = new MetadataSources(registry.build());

			// Get the database dialect and create a DbEngine
			String dialectClass = baseConfig.getProperty("hibernate.dialect");
			dialect = ClassUtil.newInstance(null, dialectClass, Dialect.class);
			dbEngine = EngineFactory.createEngine(dialect);
		}
		catch (HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}

	/**
		Second step of initialising this class. This is done at startup time by the
		{@link Application#start()} method. In this step we read all mapping files, 
		generate additional mappings for the {@link ExtendableData} items and raw data, 
		and generate filters	used by {@link Query} implementation.
	*/
	static synchronized void init2()
		throws BaseException
	{
		// Return if we have already been initialised
		if (isInitialised) return;
	
		try
		{
			// Load filter definitions from filter-def.hbm.xml
			FilterConfigurator filterConfig = new FilterConfigurator();
			filterConfig.loadFilterTemplates("/net/sf/basedb/core/filter-def.hbm.xml");
			
			addStaticMappings(cfg, filterConfig);
			addRawDataMappings(cfg);
			metadata = cfg.buildMetadata();
			
			filterConfig.registerDelayedFilters(metadata);
			
			SessionFactoryBuilder sfBuilder = metadata.getSessionFactoryBuilder();
			
			// Add database-specific functions to the configuration
			addSqlFunctions(sfBuilder, dbEngine);
			sf = sfBuilder.build();

			if (sf.getSessionFactoryOptions().getJdbcFetchSize() != null)
			{
				jdbcFetchSize = sf.getSessionFactoryOptions().getJdbcFetchSize();
			}
		}
		catch (HibernateException | IOException ex)
		{
			throw new BaseException(ex);
		}
		isInitialised = true;
	}
	
	/**
		Unload all settings.
	*/
	static synchronized void unload()
	{
		isInitialised = false;
		if (sf != null && !sf.isClosed()) sf.close();
		sf = null;
		cfg = null;
		metadata = null;
		dialect = null;
		dbEngine = null;
	}

	/**
		Read database configuration properties from the base.config file
		and map them to Hibernate settings.
		@return Properties for database connection
		@since 3.8
	*/
	private static Properties getConnectionProperties()
		throws BaseException
	{
		String dialect = Config.getString("db.dialect", "");
		if (dialect.contains("PostgreSQLDialect"))
		{
			throw new ConfigurationException("PostgreSQLDialect has been deprecated. Use PostgreSQL9Dialect instead (base.config).");
		}
		String dbUrl = Config.getString("db.url", "");
		String dbUser = Config.getString("db.username", "");
		String dbPassword = Config.getString("db.password", "");
		checkJdbcDriver(dbUrl, dbUser, dbPassword);
		
		Properties baseConfig = new Properties();
		baseConfig.setProperty("hibernate.dialect", dialect);
		baseConfig.setProperty("hibernate.connection.url", dbUrl);
		baseConfig.setProperty("hibernate.connection.username", dbUser);
		baseConfig.setProperty("hibernate.connection.password", dbPassword);
		return baseConfig;
	}

	/**
		Check that a JDBC driver is registered and try to load a driver if not.
	*/
	private static void checkJdbcDriver(String dbUrl, String dbUser, String dbPassword)
	{
		Driver driver = null;
		try
		{
			driver = DriverManager.getDriver(dbUrl);
		}
		catch (SQLException ex)
		{
			log.info("JDBC driver for '" + dbUrl + "' is not registered. Trying to re-load via ServiceLoader.");
			// The driver has not been registered which can happen when the DriverManager class
			// is initialized by the "wrong" class loader (eg. Tomcat will load this in a bootstrap
			// class loader which doesn't have access to JAR files in the WEB-INF/lib directory)
			// So, we need let the drivers register themselves via the ServiceLoader
			ServiceLoader<Driver> s = ServiceLoader.load(Driver.class);
			Iterator<Driver> it = s.iterator();
			while (driver == null && it.hasNext())
			{
				Driver d = it.next();
				if (d != null)
				{
					boolean acceptUrl = false;
					try
					{
						acceptUrl = d.acceptsURL(dbUrl);
					}
					catch (SQLException ex2)
					{}
					log.info("Loaded JDBC driver " + d.getClass().getName() + 
							"; version=" + d.getMajorVersion() + "." + d.getMinorVersion()+
							"; acceptsURL="+acceptUrl);
					if (acceptUrl) driver = d;
				}
			}
		}
		if (driver == null)
		{
			throw new ConfigurationException("Could not find a suitable JDBC driver for URL: " + dbUrl);
		}
		
		Properties props = new Properties();
		props.setProperty("user", Config.getString("db.username", ""));
		props.setProperty("password", Config.getString("db.password", ""));

		try (Connection c = driver.connect(dbUrl, props))
		{
			if (!c.isValid(5))
			{
				throw new ConfigurationException("Could not connect to database: " + dbUrl);
			}
		}
		catch (SQLException ex)
		{
			throw new ConfigurationException("Could not connect to database: " + dbUrl + " (" + ex.getMessage() + ")", ex);
		}
		
		log.info("Accepted JDBC driver " + driver.getClass().getName() + 
				"; version=" + driver.getMajorVersion() + "." + driver.getMinorVersion());
	}
	
	/**
		Read all mapping files, which must be named <code>Class.hbm.xml</code> where class
		is the name of each class. The must be located in the same directory as the class
		files, and we support loading the xml files both from a jar file of from the
		directory.
	*/
	private static void addStaticMappings(MetadataSources cfg, FilterConfigurator filterConfig)
		throws BaseException
	{
		assert cfg != null : "cfg == null";
		JarFile jarFile = null;
		try
		{
			List<URL> mappingFiles = new ArrayList<URL>();
			URL coreUrl = HibernateUtil.class.getResource("/base.version");
			String corePath = coreUrl.getPath();
			if ("jar".equals(coreUrl.getProtocol()))
			{
				// We are in a JAR file and get a path like: file:/usr/local/.../BASE2.jar!/base.version
				// Remove "file:" and "!/base.version" to get the path to the JAR
				String jarPath = corePath.substring("file:".length(), corePath.length()-"/base.version".length()-1);
				log.info("Loading mappings files from jar: " + jarPath);
				jarFile = new JarFile(jarPath);
				Enumeration<JarEntry> jarEntries = jarFile.entries();
				while (jarEntries.hasMoreElements())
				{
					JarEntry entry = jarEntries.nextElement();
					if (!entry.isDirectory() && entry.getName().endsWith(".hbm.xml"))
					{
						log.debug("Found mapping file: " + entry.getName());
						mappingFiles.add(HibernateUtil.class.getResource("/" + entry.getName()));
					}
				}
			}
			else
			{
				// We are in a directory and get a path like: /usr/local/.../base.version
				// Move up one directory to get to the root path
				java.io.File dir = new java.io.File(corePath).getParentFile();
				log.info("Loading mappings files from directory: " + dir);
				List<java.io.File> files = FileUtil.findFiles(dir, new RegexpFileFilter(".*\\.hbm\\.xml", null));
				for (java.io.File hbmFile : files)
				{
					log.debug("Found mapping file: " + hbmFile);
					mappingFiles.add(hbmFile.toURI().toURL());
				}
			}

			// Load the XML mapping files and register them with Hibernate
			// NOTE! We need to use our own copy since the
			// org/hibernate/hibernate-mapping-4.0.xsd seems to have bugs:
			//  * Only one <colum> element allowed inside <many-to-any>
			final String[] schemaFiles = {
					"http://www.hibernate.org/xsd/hibernate-mapping",
					HibernateUtil.class.getResource("/net/sf/basedb/core/xsd/hibernate-mapping-4.0.xsd").toString()
				};
			
			XMLOutputter out = new XMLOutputter();
			
			for (URL url : mappingFiles)
			{
				log.info("Loading mapping file: " + url);
				//System.out.println(url.toString());
				Document dom = XmlUtil2.getSchemaValidatedXML(url, schemaFiles);
				manipulateDocument(dom, filterConfig);
				String xml = out.outputString(dom);
				cfg.addInputStream(new ByteArrayInputStreamAccess(
					url.toString(),
					xml.getBytes(StandardCharsets.UTF_8)
					));
			}
		}
		catch (Exception ex)
		{
			log.error("Exception while loading mapping files", ex);
			throw new BaseException(ex);
		}
		finally
		{
			FileUtil.close(jarFile);
		}
	}

	/**
		Some XML files require manipulation by us before they are registered.
		<ul>
		<li>Add extended properties to extendable classes
		<li>Add filter implementation to almost all classes
		<li>Other special-case fixes depending on the underlying database
		</ul>
		
	*/
	private static void manipulateDocument(Document dom, FilterConfigurator filterConfig)
	{
		Element root = dom.getRootElement();
		
		// Load all <class> and <subclass> tags in the document
		Iterator<Element> it = root.getDescendants(new org.jdom2.filter.AbstractFilter<Element>()
			{
				private static final long serialVersionUID = 1L;

				@Override
				public Element filter(Object obj)
				{
					if (!(obj instanceof Element)) return null;
	
					Element e = (Element)obj;
					String name = e.getName();
					return "class".equals(name) || "subclass".equals(name) ? e : null;
				}
			});
		
		// Copy to a list since we can't use the iterater and add elements to the
		// document at the same time (ConcurrentModificationException)
		List<Element> classTags = new ArrayList<Element>();
		while (it.hasNext())
		{
			classTags.add(it.next());
		}
		
		for (Element classTag: classTags)
		{
			String className = classTag.getAttributeValue("name");
		
			// Extended properties
			if (ExtendedProperties.isExtendable(className))
			{
				addExtendedProperties(classTag, className);
			}
		
			// Permission filters
			addFilterConditions(classTag, className, filterConfig);
			
			// Unit symbol need to be case-sensitive
			if (className.endsWith("UnitSymbolData"))
			{
				fixUnitSymbolColumnDefinition(classTag);
			}
		}
	}
	
	/**
		Append &lt;property&gt; tags to the given class tag.
		@param classTag An Element representing a &lt;class&gt; tag
		@param className The name of the ({@link ExtendableData}) class
		@since 2.17
	*/
	private static void addExtendedProperties(Element classTag, String className)
	{
		log.info("Adding extended properties to class: " + className);
		List<ExtendedProperty> properties = ExtendedProperties.getProperties(className, true);

		for (ExtendedProperty property : properties)
		{
			log.debug("Adding extended property to class '" + className + "': " + property.getName());
			classTag.addContent(createPropertyElement(property, classTag.getNamespace()));
		}
	}
	
	/**
		Creates a &lt;property&gt; tag from an extended property
		definition. This also works for raw data properties since they
		extend from ExtendedProperty.
		@since 2.17
	*/
	private static Element createPropertyElement(ExtendedProperty property, Namespace ns)
	{
		Element element = new Element("property", ns);
		element.setAttribute("name", property.getName());
		element.setAttribute("column", "`"+property.getColumn()+"`");
		element.setAttribute("type", property.getType().getStringValue());
		element.setAttribute("length", Integer.toString(property.getLength()));
		element.setAttribute("not-null", Boolean.toString(!property.isNullable()));
		element.setAttribute("access", "net.sf.basedb.core.ExtendedPropertyAccessor");
		return element;
	}
	
	/**
		Add &lt;filter&gt; conditions to the given class tag.
		@param classTag An Element representing a &lt;class&gt; tag
		@param className The name of the entity class
		@since 2.17
	*/
	private static void addFilterConditions(Element classTag, String className, FilterConfigurator filterConfig)
	{
		try
		{
			Class<?> c = Class.forName(className);
			Item itemType = Item.fromClass(c);

			log.info("Adding 'denyAll' filter to "+c.getName());
			defineFilter(classTag, itemType, "denyAll", "`id` != `id`", filterConfig);
			if (NewsData.class.isAssignableFrom(c))
			{
				log.info("Adding 'todaysNews' filter to "+c.getName());
				defineFilter(classTag, itemType, "todaysNews", 
						":today >= `start_date` AND (:today <= `end_date` OR `end_date` IS NULL)", filterConfig);
			}
			if (MessageData.class.isAssignableFrom(c))
			{
				log.info("Adding 'ownedBy' filter to " + c.getName());
				defineFilter(classTag, itemType, "ownedBy", ":owner = `to_user_id`", filterConfig);
				defineFilter(classTag, itemType, "ownedOrRemovedBy", "(:user = `to_user_id` OR :user = `removed_by`)", filterConfig);
			}
			if (SessionData.class.isAssignableFrom(c))
			{
				log.info("Adding 'ownedBy' filter to " + c.getName());
				defineFilter(classTag, itemType, "ownedBy", ":owner = `user_id`", filterConfig);
			}
			if (UserDeviceData.class.isAssignableFrom(c))
			{
				log.info("Adding 'ownedBy' filter to " + c.getName());
				defineFilter(classTag, itemType, "ownedBy", ":owner = `user_id`", filterConfig);
			}
			if (AnnotationData.class.isAssignableFrom(c))
			{
				log.info("Adding 'annotationTypes' filter to " + c.getName());
				defineFilter(classTag, itemType, "annotationTypes", "`annotationtype_id` IN (:annotationTypes)", filterConfig);
			}
			if (RemovableData.class.isAssignableFrom(c))
			{
				log.info("Adding 'isRemoved' filter to "+c.getName());
				defineFilter(classTag, itemType, "isRemoved", ":removed <> (`removed_by` IS NULL)", filterConfig);
				if (OwnableData.class.isAssignableFrom(c))
				{
					defineFilter(classTag, itemType, "ownedOrRemovedBy", "(:user = `owner` OR :user = `removed_by`)", filterConfig);
				}
			}
			if (AnnotatableData.class.isAssignableFrom(c))
			{
				log.info("Adding 'isAnnotatated' filter to "+c.getName());
				defineFilter(classTag, itemType, "isAnnotated", ":annotated = (`annotationset_id` IS NOT NULL)", filterConfig);
			}
			if (OwnableData.class.isAssignableFrom(c))
			{
				log.info("Adding 'ownedBy' filter to "+c.getName());
				defineFilter(classTag, itemType, "ownedBy", ":owner = `owner`", filterConfig);
				
				log.info("Adding 'notOwnedBy' filter to "+c.getName());
				defineFilter(classTag, itemType, "notOwnedBy", ":owner != `owner`", filterConfig);
	
				log.info("Adding 'memberOf' filter to "+c.getName());
				defineFilter(classTag, itemType, "memberOf", ":owner != `owner` AND `id` IN (:items)", filterConfig);
	
				log.info("Adding 'ownedByOrMemberOf' filter to "+c.getName());
				defineFilter(classTag, itemType, "ownedByOrMemberOf", "(:owner = `owner` OR `id` IN (:items))", filterConfig);
			}
			else if (UserData.class.isAssignableFrom(c))
			{
				log.info("Adding 'memberOf' filter to "+c.getName());
				defineFilter(classTag, itemType, "memberOf", "(`id` IN (:items) OR `id` = :owner)", filterConfig);
			}
			else
			{
				log.info("Adding 'memberOf' filter to "+c.getName());
				defineFilter(classTag, itemType, "memberOf", "`id` IN (:items)", filterConfig);
			}
			if (ShareableData.class.isAssignableFrom(c))
			{
				log.info("Adding 'sharedTo' filter to "+c.getName());
				defineFilter(classTag, itemType, "sharedTo", ":owner != `owner` AND `itemkey_id` IN (:itemKeys)", filterConfig);
	
				log.info("Adding 'inProject' filter to "+c.getName());
				defineFilter(classTag, itemType, "inProject", "`projectkey_id` IN (:projectKeys)", filterConfig);
	
				log.info("Adding 'ownedByOrSharedTo' filter to "+c.getName());
				defineFilter(classTag, itemType, "ownedByOrSharedTo", "(:owner = `owner` OR `itemkey_id` IN (:itemKeys))", filterConfig);
	
				log.info("Adding 'ownedByOrInProject' filter to "+c.getName());
				defineFilter(classTag, itemType, "ownedByOrInProject", 
						"(:owner = `owner` OR `projectkey_id` IN (:projectKeys))", filterConfig);
	
				log.info("Adding 'sharedToOrInProject' filter to "+c.getName());
				defineFilter(classTag, itemType, "sharedToOrInProject", 
						"((:owner != `owner` AND `itemkey_id` IN (:itemKeys)) OR `projectkey_id` IN (:projectKeys))", filterConfig);
	
				log.info("Adding 'ownedByOrSharedToOrInProject' filter to "+c.getName());
				defineFilter(classTag, itemType, "ownedByOrSharedToOrInProject", 
						"(:owner = `owner` OR `itemkey_id` IN (:itemKeys) OR `projectkey_id` IN (:projectKeys))", filterConfig);
			}
		}
		catch (ClassNotFoundException ex)
		{
			throw new RuntimeException(ex);
		}
	}
		
	/**
		Define a filter for on the given class.
		@since 3.6
	*/
	private static void defineFilter(Element classTag, Item itemType, String name, String condition, FilterConfigurator filterConfig)
	{
		// Generic filter condition is only used for '<class>' elements (not '<subclass>')
		if (classTag.getName().equals("class"))
		{
			filterConfig.createGenericFilter(name, condition, classTag);
		}
		
		// Define item-specific filters with name+itemType so we have
		// control for each type of item
		if (itemType != null)
		{
			filterConfig.createItemSpecificFilterDef(name, itemType, classTag);
			filterConfig.createItemSpecificFilter(name, itemType, condition, classTag);
		}
	}
	
	/**
		Generate additional mappings for the raw data types.
	*/
	private static void addRawDataMappings(MetadataSources cfg)
		throws BaseException
	{
		final String[] schemaFiles = {
				"http://www.hibernate.org/xsd/hibernate-mapping",
				HibernateUtil.class.getResource("/net/sf/basedb/core/xsd/hibernate-mapping-4.0.xsd").toString()
			};
		
		// The template XML mapping file
		final URL templateFile = HibernateUtil.class.getResource("/net/sf/basedb/core/templates/hibernate-properties-RawData.xml");

		XMLOutputter out = new XMLOutputter();
		for (RawDataType rdt : RawDataTypes.getRawDataTypes() )
		{
			if (rdt.isStoredInDb())
			{
				String entityName = rdt.getEntityName();
				String tableName = rdt.getTableName();
				log.info("Adding raw data type: " + entityName);
				try
				{
					Document dom = XmlUtil2.getSchemaValidatedXML(templateFile, schemaFiles);
					Namespace ns = dom.getRootElement().getNamespace();
					Element clazz = dom.getRootElement().getChild("class", ns);
					// Specify an entity name and database table for the raw data type
					clazz.setAttribute("entity-name", entityName);
					clazz.setAttribute("table", "`"+tableName+"`");
	
					// Specify a sequence name for id generator
					Element generator = clazz.getChild("id", ns).getChild("generator", ns);
					Element sequenceName = new Element("param", ns);
					sequenceName.setAttribute("name", "sequence_name");
					sequenceName.setText(entityName+"_sequence");
					generator.addContent(sequenceName);

					// Add admin-defnied properties
					for (RawDataProperty property : rdt.getProperties())
					{
						log.info("Adding raw data property to class '" + entityName + "': " + property.getName());
						clazz.addContent(createPropertyElement(property, ns));
					}
					cfg.addInputStream(new ByteArrayInputStreamAccess(
						templateFile.toString()+"?"+rdt.getEntityName(),
						out.outputString(dom).getBytes(StandardCharsets.UTF_8)));
				}
				catch (Exception ex)
				{
					log.error("Exception while adding raw data type: " + entityName, ex);
					throw new BaseException(ex);
				}
				log.info("Added mapping for raw data type: "+entityName);
			}
		}
	}

	/**
		Add more SQL functions to HQL that are not present in a 
		standard Hibernate installation. SQL functions are registered
		on the configuration with 
		{@link Configuration#addSqlFunction(String, SQLFunction)}. Note!
		The name of the function must be in lower-case letters, otherwise
		Hibernate will not be able to find it. In HQL, it is possible to use
		both lower-case and uppercase. 
		<p>
		Here are the functions registered so far.
		
		<ul>
		<li>rlike: Match a string against a regular expression
		</ul>
		
		@since 2.8, 3.8
	*/
	private static void addSqlFunctions(SessionFactoryBuilder sfBuilder, DbEngine dbEngine)
	{
		sfBuilder.applySqlFunction("rlike", 
			new SQLFunctionTemplate(TypeWrapper.BOOLEAN.getHibernateType(), "(" + dbEngine.rlike("?1", "?2") + ")"));
		sfBuilder.applySqlFunction("native", new NativeSQLFunction());
	}
	
	/**
		Fixes the definition of the UnitSymbols.symbol column so that
		it is case-sensitive. We need because some units have both
		milli (m) and Mega (M) as prefix (eg. mV and MV)
		@since 2.17
	*/
	private static void fixUnitSymbolColumnDefinition(Element classTag)
	{
		// Make the UnitSymbols.symbol column case-sensitive so we can have units
		// with both milli (m) and Mega (M) as prefix.
		if (!dbEngine.caseInsensitiveComparison()) return;
		log.debug("Setting UnitSymbols.symbol to a case-sensitive column type");
		
		// Find the class/property/column[name=`symbol`] tag
		Iterator<Element> match = classTag.getDescendants(
			new org.jdom2.filter.AbstractFilter<Element>()
			{
				private static final long serialVersionUID = -2876632679035736428L;
				@Override
				public Element filter(Object obj)
				{
					if (!(obj instanceof Element)) return null;
	
					Element e = (Element)obj;
					return e.getName().equals("column") && e.getAttributeValue("name").equals("`symbol`") ? e : null;
				}
			});
		Element col = match.next();
		String sqlType = dbEngine.getCaseSensitiveVarchar(XmlUtil2.getIntAttribute(col, "length", 255));
		if (sqlType != null) 
		{
			col.setAttribute("sql-type", sqlType);
		}
	}
	
	/**
		Create the all tables in the database. The database must exists and we 
		recommend that it is empty to begin with.
		@return The schemaVersion that is currently installed
	*/
	static int createStaticTables(SchemaGenerator.Mode mode, ProgressReporter progress)
		throws BaseException
	{
		assert sf != null : "HibernateUtil has not been initialised";

		int currentSchemaVersion = Install.NEW_SCHEMA_VERSION;
		if (mode.verifySchemaVersion())
		{
			currentSchemaVersion = Application.getSchemaVersion();				
			if (currentSchemaVersion < 0)
			{
				if (currentSchemaVersion == -1 && Install.NEW_SCHEMA_VERSION == 118)
				{
					// Special update that adds a column to SchemaVersion table causing this check to fail!
					// But we allow the update to continue with the schema version from BASE 3.3
					currentSchemaVersion = 117;
				}
				else
				{
					throw new BaseException("Couldn't load current schema version of database");
				}
			}
			else if (Install.NEW_SCHEMA_VERSION < currentSchemaVersion)
			{
				throw new BaseException("Can't update to schema version " + Install.NEW_SCHEMA_VERSION + 
						" of database, cause you already have schema version " + currentSchemaVersion + " installed");
			}
			else if (currentSchemaVersion < 100)
			{
				throw new BaseException("Updating from BASE 2.17 is no longer supported. Please update to BASE 3.3 or earlier first.");
			}
		}
		
		boolean emptyDatabase = isEmptyDatabase();
		if (mode.requireEmptyDatabase() && !emptyDatabase)
		{
			throw new BaseException("Database already has tables.");
		}
		else if (!mode.requireEmptyDatabase() && emptyDatabase)
		{
			throw new BaseException("Database is empty");
		}

		Session session = null;
		Transaction tx = null;
		try
		{
			SchemaGenerator schemaGenerator = new SchemaGenerator(metadata, dialect, dbEngine, mode, progress);
			session = HibernateUtil.newSession();
			tx = HibernateUtil.newTransaction(session);
			session.doWork(schemaGenerator);
			HibernateUtil.commit(tx);
		}
		catch (Exception ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			throw new BaseException(ex);
		}
		finally
		{
			if (session != null) HibernateUtil.close(session);
		}
		return currentSchemaVersion;
	}

	/**
		Check if the main database has any tables.
		@return TRUE if there are any tables, FALSE otherwise
	*/
	static boolean isEmptyDatabase()
		throws BaseException
	{
		Session session = null;
		boolean isEmpty = true;
		try
		{
			session = HibernateUtil.newSession();
			isEmpty = HibernateUtil.doReturningWork(session, 
					new ReturningWork<Boolean>()
					{
						@Override
						public Boolean execute(Connection connection)
							throws SQLException
						{
							boolean isEmpty = true;
							DatabaseMetaData metaData = connection.getMetaData();
							String catalog = connection.getCatalog();
							String schema = connection.getSchema();
							Iterator<PersistentClass> classes = metadata.getEntityBindings().iterator();
							while (classes.hasNext() && isEmpty)
							{
								PersistentClass pClass = classes.next();
								Table t = pClass.getTable();
								ResultSet tables = metaData.getTables(catalog, schema, t.getName(), new String[] { "TABLE" } );
								if (tables.next())
								{
									isEmpty = false;
									log.debug("Table '"+t.getName()+"' already exists");
								}
								else
								{
									log.debug("Table '"+t.getName()+"' doesn't exists");
								}
								tables.close();
							}
							return isEmpty;
						}
					}
				);
			
			HibernateUtil.close(session);
			session = null;
		}
		catch (SQLException ex)
		{
			throw new BaseException(ex);
		}
		catch (HibernateException ex)
		{
			throw new BaseException(ex);
		}
		finally
		{
			if (session != null) HibernateUtil.close(session);
		}
		return isEmpty;
	}
	
	/**
		Create a table in the dynamic database. The real table name is
		given by the {@link VirtualTable#getTableName(VirtualDb)}. It is assumed
		that the table doesn't already exists. This can be checked by 
		{@link #virtualTableExists(VirtualDb, VirtualTable)}. This method
		uses a separate connection to the database because some databases (ie. MySQL) 
		can't handle transactions properly for this type of SQL statements.
		<p>
		This method uses classes in the <code>org.hibernate.mapping</code> package
		to define a Hibernate representation of the table and uses the 
		<code>Table.sqlCreateString</code> method to generate the SQL. This should
		ensure that this code works with all database engines supported by Hibernate.
		
		@param db The virtual db item
		@param table Information about the table to create
		@return The name of the created table
		@see #dropVirtualTable(VirtualDb, VirtualTable)
		@see #virtualTableExists(VirtualDb, VirtualTable)
	*/
	static synchronized String createVirtualTable(VirtualDb db, VirtualTable table)
		throws BaseException
	{
		Session session = null;
		Transaction tx = null;
		try
		{
			session = HibernateUtil.newSession();
			tx = HibernateUtil.newTransaction(session);
			String tableName = table.getTableName(db);

			if (!tableExists(tableName, session))
			{
				Table hibernateTable = newDynamicMapping(db, table, true);
				Dialect dialect = getDialect();
				String dynamicCatalog = Application.getDynamicCatalog();
				String quotedCatalog = quote(dynamicCatalog);
				String dynamicSchema = Application.getDynamicSchema();
				String quotedSchema = quote(dynamicSchema);

				Database database = metadata.getDatabase();
				SqlStringGenerationContext sqlContext = SqlStringGenerationContextImpl.fromExplicit(
					database.getJdbcEnvironment(), database, quotedCatalog, quotedSchema);
				List<String> statements = new ArrayList<String>();

				String sql = hibernateTable.sqlCreateString(metadata, sqlContext, quotedCatalog, quotedSchema);
				sql = dbEngine.makeSafeCreateTable(sql, dynamicCatalog, dynamicSchema, tableName);
				logSql.debug(sql);
				statements.add(sql);
				
				Iterator<Index> i = hibernateTable.getIndexIterator();
				while (i.hasNext())
				{
					Index ii = i.next();
					sql = ii.sqlCreateString(metadata, sqlContext, quotedCatalog, quotedSchema);
					logSql.debug(sql);
					statements.add(sql);
				}
				HibernateUtil.doWork(session, new MultiUpdateWork(statements));
			}
			HibernateUtil.commit(tx);
			return tableName;
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			log.error("Exception while creating virtual table: " + table, ex);
			throw ex;
		}
		catch (SQLException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			log.error("Exception while creating virtual table: " + table, ex);
			throw new BaseException(ex);
		}
		finally
		{
			if (session != null) HibernateUtil.close(session);
		}
	}
	
	/**
		Drop a table in the dynamic database. The real table name is
		given by the {@link VirtualTable#getTableName(VirtualDb)}. If the table
		doesn't exists this method does nothing. This method
		uses a separate connection to the database because some databases (ie. MySQL) 
		can't handle transactions properly for this type of SQL statements.
		<p>
		This method uses classes in the <code>org.hibernate.mapping</code> package
		to define a Hibernate representation of the table and uses the 
		<code>Table.sqlDropString</code> method to generate the SQL. This should
		ensure that this code works with all database engines supported by Hibernate.
		
		@param db The virtual db item
		@param table Information about the table to drop
		@see #createVirtualTable(VirtualDb, VirtualTable)
		@see #virtualTableExists(VirtualDb, VirtualTable)
	*/
	static synchronized void dropVirtualTable(VirtualDb db, VirtualTable table)
		throws BaseException
	{
		dropDynamicTable(table.getTableName(db));
	}
	
	static synchronized void dropDynamicTable(String tableName)
	{
		Session session = null;
		Transaction tx = null;
		try
		{
			session = HibernateUtil.newSession();
			tx = HibernateUtil.newTransaction(session);

			if (tableExists(tableName, session))
			{
				Table hibernateTable = new Table("`" + tableName + "`");
				Dialect dialect = getDialect();
				String dynamicCatalog = quote(Application.getDynamicCatalog());
				String dynamicSchema = quote(Application.getDynamicSchema());
				Database database = metadata.getDatabase();
				SqlStringGenerationContext sqlContext = SqlStringGenerationContextImpl.fromExplicit(
					database.getJdbcEnvironment(), database, dynamicCatalog, dynamicSchema);

				String sql = hibernateTable.sqlDropString(sqlContext, dynamicCatalog, dynamicSchema);
				logSql.debug(sql);
				HibernateUtil.doReturningWork(session, new ExecuteUpdateWork(null, sql));
			}
			HibernateUtil.commit(tx);
		}
		catch (SQLException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			log.error("Exception while dropping dynamic table: " + tableName, ex);
			throw new BaseException(ex);
		}
		catch (RuntimeException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			log.error("Exception while dropping dynamic table: " + tableName, ex);
			throw ex;
		}
		finally
		{
			if (session != null) HibernateUtil.close(session);
		}
	}
	
	/**
		Check if a table exists in the dynamic database or not. The real table 
		name is given by the {@link VirtualTable#getTableName(VirtualDb)}. 
		<p>
		This method uses JDBC {@link DatabaseMetaData} object on a
		separate connection to find out if the table exists or not. 
		
		@param db The virtual db item
		@param table Information about the table
		@see #createVirtualTable(VirtualDb, VirtualTable)
		@see #dropVirtualTable(VirtualDb, VirtualTable)
	*/
	static synchronized boolean virtualTableExists(VirtualDb db, VirtualTable table)
	{
		Session session = null;
		Transaction tx = null;
		boolean hasTable = false;
		try
		{
			session = HibernateUtil.newSession();
			tx = HibernateUtil.newTransaction(session);
			hasTable = tableExists(table.getTableName(db), session);
			HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			log.error("Exception while checking virtual table: " + table, ex);
			throw ex;
		}
		catch (SQLException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			log.error("Exception while checking virtual table: " + table, ex);
			throw new BaseException(ex);
		}
		finally
		{
			if (session != null) HibernateUtil.close(session);
		}
		return hasTable;
	}
	
	/**
		Check if table in the dynamic database exists.
		@param tableName The name of the table
		@param session The Hibernate session
		@see #virtualTableExists(VirtualDb, VirtualTable)
	*/
	private static boolean tableExists(String tableName, Session session)
		throws SQLException
	{
		return HibernateUtil.doReturningWork(session, 
			new TableExistsWork(Application.getDynamicCatalog(), Application.getDynamicSchema(), tableName));
	}
	
	/**
		Create a new <code>org.hibernate.mapping.Table</code> object for
		the specified table. This method is used by 
		{@link #createVirtualTable(VirtualDb, VirtualTable)} to build the mapping
		information Hibernate requires to generate the SQL to create the table.
		@param db The virtual db item
		@param table Information about the table to generate the mapping for
	*/
	private static Table newDynamicMapping(VirtualDb db, VirtualTable table, boolean withColumns)
	{
		String tableName = table.getTableName(db);
		Table hibernateTable = new Table("`"+tableName+"`");
		if (withColumns)
		{
			hibernateTable.setPrimaryKey(new PrimaryKey(hibernateTable));
			for (VirtualColumn column : table.getColumns(db))
			{
				addColumn((org.hibernate.boot.spi.MetadataImplementor)metadata, hibernateTable, column);
			}
		}
		return hibernateTable;
	}
	
	/**
		Add a column to a table. This method is used by 
		{@link #newDynamicMapping(VirtualDb, VirtualTable, boolean)} to build the mapping
		information Hibernate requires to generate the SQL to create the table.
		@param hibernateTable The <code>org.hibernate.mapping.Table</code> object
			that holds the mapping for the table
		@param column The column to add to the table
	*/
	private static void addColumn(org.hibernate.boot.spi.MetadataImplementor metadata, Table hibernateTable, VirtualColumn column)
	{
		// Create Hibernate Column mapping
		Column hibernateColumn = new Column("`"+column.getColumn()+"`");
		hibernateColumn.setSqlTypeCode(column.getTypeWrapper().getSQLType());
		hibernateColumn.setNullable(column.isNullable());
		hibernateColumn.setLength(column.getSize());
		hibernateColumn.setSqlType(dialect.getTypeName(hibernateColumn.getSqlTypeCode(), hibernateColumn.getLength(), hibernateColumn.getPrecision(), hibernateColumn.getScale()));
		
		// Add column to Table mapping
		hibernateTable.addColumn(hibernateColumn);
		
		if (column.isPrimaryKey()) 
		{
			// Add column to primary key
			hibernateTable.getPrimaryKey().addColumn(hibernateColumn);
		}
		else if (column.isUnique())
		{
			// Set unique constraint for column
			hibernateColumn.setUnique(column.isUnique());
		}
		else if (column.isIndexed()) 
		{
			// Create index on this column
			Index hibernateIndex = new Index();
			hibernateIndex.addColumn(hibernateColumn);
			hibernateIndex.setName(quote(hibernateTable.getName()+column.getName()));
			hibernateIndex.setTable(hibernateTable);
			hibernateTable.addIndex(hibernateIndex);
		}
	}
	
	static Connection getPlainConnection()
		throws SQLException
	{
		String url = Config.getString("db.url");
		String user = Config.getString("db.username");
		String pwd = Config.getString("db.password");
		Connection conn = DriverManager.getConnection(url, user, pwd);
		return conn;
	}
	
	/**
		Create a new Hibernate session. 
	*/
	static Session newSession()
		throws BaseException
	{
		return newSession(null, null);
	}
	
	static Session newSession(Connection c)
	{
		return newSession(null, c);
	}
	
	/**
		Create a new Hibernate session, optionally with an interceptor.
		@since 2.13
	*/
	static Session newSession(Interceptor interceptor, Connection c)
		throws BaseException
	{
		assert sf != null : "HibernateUtil has not been initialised";
		try
		{
			SessionBuilder<?> builder = sf.withOptions();
			if (interceptor != null) builder.interceptor(interceptor);
			if (c != null) builder.connection(c);
			Session session = builder.openSession();
			// FlushMode.COMMIT means that all changes
			// are automatically flushed before a commit is performed
			session.setHibernateFlushMode(FlushMode.COMMIT);
			/*
				If you get an exception for setTransactionIsolation it may be a problem
				of a pooled connection that wasn't commited or rollbacked by the previous
				user. The following command can fix it: 
				
				session.connection().rollback()
				==================================================================
				BUT!!! TRY TO FIND THE PLACE WHERE THE TRANSACTION IS LEFT OPEN!!!
				==================================================================
			*/
			// Important: we may get into problem with Keys otherwise which do their work
			// in separate transactions
			HibernateUtil.doWork(session, 
					new Work()
					{
						@Override
						public void execute(Connection connection) 
							throws SQLException
						{
							connection.setTransactionIsolation(Connection.TRANSACTION_READ_COMMITTED);
						}
					}
			);
			return session;
		}
		catch(HibernateException ex)
		{
			throw new BaseException(ex);
		}
		catch (SQLException ex)
		{
			throw new BaseException(ex);
		}
	}

	static StatelessSession newStatelessSession(Session session)
		throws BaseException
	{
		assert sf != null : "HibernateUtil has not been initialised";
		assert session != null : "session == null";
		try
		{
			return sf.openStatelessSession(HibernateUtil.getConnection(session));
		}
		catch (HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}
	
	/**
		Create a new transaction for a session. If a transaction has already been created 
		for the session, that one is returned.
	*/
	static Transaction newTransaction(Session session)
		throws BaseException
	{
		assert session != null : "session == null";
		try
		{
			return session.beginTransaction();
		}
		catch(HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}

	/**
		Get the underlying JDBC connection from the Hibernate session.
		Hibernate 3.3 have implemented a partial replacement for the
		Session.connection() method that can be used if the connection
		or any derived object such as a PreparedStatement or ResultSet is
		not needed after the connection has been used. Such code should
		use {@link #doWork(Session, Work)} instead.
		<p>
		Other code may continue using this method until the Hibernate
		team develops a full replacement.
		See 
		<a href="http://forum.hibernate.org/viewtopic.php?t=974518">
		http://forum.hibernate.org/viewtopic.php?t=974518</a>
		@since 2.4
		@see #doWork(Session, Work)
	*/
	static Connection getConnection(Session session)
	{
		try
		{
			if (session instanceof SessionWrapper)
			{
				session = ((SessionWrapper)session).getParentSession();
			}
			return ((SessionImpl)session).connection();
		}
		catch (HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}
	
	/**
		Get the configured session factory.
		@since 2.15
	*/
	static SessionFactory getSessionFactory()
	{
		return sf;
	}
	
	/**
		Execute some arbitrary JDBC code using the same
		database connection as the Hibernate session. This method
		should be used instead of {@link #getConnection(Session)} 
		wherever possible.

		@param session The Hibernate session to use
		@param work The work implementation
		@throws SQLException In case there is an 
			SQL error
		@since 3.4
	*/
	static void doWork(Session session, Work work)
		throws SQLException
	{
		try
		{
			session.doWork(work);
		}
		catch (HibernateException ex)
		{
			if (ex.getCause() instanceof SQLException)
			{
				throw (SQLException)ex.getCause();
			}
			throw ex;
		}
	}
	
	/**
		Execute some arbitrary JDBC code using the same
		database connection as the Hibernate session. This method
		should be used instead of {@link #getConnection(Session)} 
		wherever possible.
	
		@param session The Hibernate session to use
		@param work The work implementation
		@return Determined by the work implementation
		@throws SQLException In case there is an 
			SQL error
		@since 3.4
	*/
	static <R> R doReturningWork(Session session, ReturningWork<R> work)
		throws SQLException
	{
		R result = null;
		try
		{
			result = session.doReturningWork(work);
		}
		catch (HibernateException ex)
		{
			if (ex.getCause() instanceof SQLException)
			{
				throw (SQLException)ex.getCause();
			}
			throw ex;
		}
		return result;
	}
	
	/**
		Commit a transaction.
	*/
	static void commit(Transaction tx)
		throws BaseException
	{
		assert tx != null : "tx == null";
		try
		{
			tx.commit();
		}
		catch (PersistenceException ex)
		{
			Throwable cause = ex.getCause();
			if (cause instanceof HibernateException) ex = (HibernateException)cause;
			
			if (ex instanceof ConstraintViolationException)
			{
				// This is the best we can do... I think...
				throw new DatabaseException(((ConstraintViolationException)ex).getSQLException());
			}
			else if (ex instanceof StaleStateException)
			{
				throw new ItemModifiedException((StaleStateException)ex);
			}
			else
			{
				throw new BaseException(ex);
			}
		}
	}
	
	/**
		Rollback a transaction.
	*/
	static void rollback(Transaction tx)
		throws BaseException
	{
		assert tx != null : "tx == null";
		try
		{
			tx.rollback();
		}
		catch (RuntimeException ex)
		{
			log.error("Exception when rolling back transaction", ex);
		}
	}

	/**
		Close a session.
	*/
	static void close(Session session)
	{
		assert session != null : "session == null";
		try
		{
			if (session.isOpen()) session.close();
		}
		catch (RuntimeException ex)
		{
			log.error("Exception when closing session", ex);
		}
	}
	
	/**
		Close a statless session.
	*/
	static void close(StatelessSession session)
	{
		assert session != null : "session == null";
		try
		{
			session.close();
		}
		catch (RuntimeException ex)
		{
			log.error("Exception when closing session", ex);
		}
	}
	
	/**
		Flush all changed objects to the database.
		@see #clear(Session)
	*/
	static void flush(Session session)
		throws BaseException
	{
		assert session != null : "session == null";
		try
		{
			session.flush();
		}
		catch (PersistenceException ex)
		{
			Throwable cause = ex.getCause();
			if (cause instanceof HibernateException) ex = (HibernateException)cause;
			
			if (ex instanceof ConstraintViolationException)
			{
				// This is the best we can do... I think...
				throw new DatabaseException(((ConstraintViolationException)ex).getSQLException());
			}
			else if (ex instanceof StaleStateException)
			{
				throw new ItemModifiedException((StaleStateException)ex);
			}
			else
			{
				throw new BaseException(ex);
			}
		}
	}

	/**
		Clear the session of cached objects. Changes that has not been written
		to the database are lost.
		@see #flush(Session)
	*/
	static void clear(Session session)
		throws BaseException
	{
		assert session != null : "session == null";
		try
		{
			session.clear();
		}
		catch(HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}
	
	/**
		Save a new data object to the database.
	*/
	static void saveData(Session session, BasicData data)
		throws BaseException
	{
		assert session != null : "session == null";
		assert data != null : "data == null";
		try
		{
			session.save(data);
		}
		catch (PersistenceException ex)
		{
			Throwable cause = ex.getCause();
			if (cause instanceof HibernateException) ex = (HibernateException)cause;
			
			if (ex instanceof ConstraintViolationException)
			{
				// This is the best we can do... I think...
				throw new DatabaseException(((ConstraintViolationException)ex).getSQLException());
			}
			else
			{
				throw new BaseException(ex);
			}
		}
	}

	/**
		Save a new data object to the database.
	*/
	static void saveData(StatelessSession session, BasicData data)
		throws BaseException
	{
		assert session != null : "session == null";
		assert data != null : "data == null";
		try
		{
			session.insert(data);
		}
		catch (PersistenceException ex)
		{
			Throwable cause = ex.getCause();
			if (cause instanceof HibernateException) ex = (HibernateException)cause;
			
			if (ex instanceof ConstraintViolationException)
			{
				// This is the best we can do... I think...
				throw new DatabaseException(((ConstraintViolationException)ex).getSQLException());
			}
			else
			{
				throw new BaseException(ex);
			}
		}
	}

	/**
		Update an existing data object in the database.
	*/
	static void updateData(Session session, BasicData data)
		throws BaseException
	{
		assert session != null : "session == null";
		assert data != null : "data == null";
		try
		{
			session.update(data);
		}
		catch(HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}

	/**
		Reconnect a disconnected data object to the session, discarding changes
		that have been made while the object was disconnected. This method is
		most useful with <code>lockOptions</code> = {@link LockOptions#NONE} if
		the logged in user only has read permission to the object.
		@see DbControl#reattachItem(BasicItem, boolean)
	*/
	static void lockData(Session session, BasicData data, LockOptions lockOptions)
		throws BaseException
	{
		assert session != null : "session == null";
		assert data != null : "data == null";
		assert lockOptions != null : "lockOptions == null";
		try
		{
			session.buildLockRequest(lockOptions).lock(data);
		}
		catch(HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}

	/**
		Load a data item from the database when you know the id.
		@param session The Hibernate session which is connected to the
			database
		@param clazz An object of this class is returned with the data
		@param id The id of the item to load
	*/
	static <T> T loadData(Session session, Class<T> clazz, int id)
		throws BaseException
	{
		assert session != null : "session == null";
		try
		{
			return clazz.cast(session.get(clazz, id));
		}
		catch(HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}

	/**
		Load a data item from the database when you know the id.
		@param session The Hibernate session which is connected 
			to the database
		@param clazz An object of this class is returned with the data
		@param entityName The entity name of the item
		@param id The id of the item to load
		@since 2.15
	*/
	static <T> T loadData(Session session, Class<T> clazz, String entityName, int id)
		throws BaseException
	{
		assert session != null : "session == null";
		try
		{
			return clazz.cast(session.get(entityName, id));
		}
		catch(HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}
	
	/**
		Reload the data from the database for a given entity.
		@param session The Hibernate session which is connected to the
			database
		@param data The entity to reload
		@since 2.4
	*/
	static void refresh(Session session, BasicData data)
	{
		assert session != null : "session == null";
		try
		{
			session.refresh(data);
		}
		catch(HibernateException ex)
		{
			throw new BaseException(ex);
		}		
	}
	
	/**
		If the given data object is a proxy, load the real data object instead.
		@since 3.3.3
	*/
	static BasicData unwrapIfProxy(Session session, BasicData data)
	{
		if (data instanceof org.hibernate.proxy.HibernateProxy)
		{
			org.hibernate.proxy.HibernateProxy proxy = (org.hibernate.proxy.HibernateProxy)data;
			org.hibernate.proxy.LazyInitializer li = proxy.getHibernateLazyInitializer();
			if (li.isUninitialized())
			{
				if (session instanceof SessionWrapper)
				{
					session = ((SessionWrapper)session).getParentSession();
				}
				li.setSession((org.hibernate.internal.SessionImpl)session);
			}
			data = (BasicData)li.getImplementation();
		}

		return data;
	}
	
	/**
		Checks if an item with the specified ID exists in the database.
		@param session The Hibernate session which is connected to the
			database
		@param clazz The type of the item
		@param id The ID of the item
		@return TRUE if the item exists in the database, FALSE otherwise
		@since 2.4
	*/
	static boolean exists(Session session, Class<?> clazz, int id)
	{
		try
		{
			String hql = "SELECT COUNT(*)"+
				" FROM "+clazz.getName()+" item"+
				" WHERE item.id = " + id;
			Query<Long> query = session.createQuery(hql, Long.class);
			return query.uniqueResult() != 0;
		}
		catch (HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}
	
	/**
		Load a data item from the database using a statless session 
		when you know the id.
		@param session The Hibernate stateless session which is connected 
			to the database
		@param clazz An object of this class is returned with the data
		@param id The id of the item to load
	*/
	static <T> T loadData(StatelessSession session, Class<T> clazz, int id)
		throws BaseException
	{
		assert session != null : "session == null";
		try
		{
			return clazz.cast(session.get(clazz, id));
		}
		catch(HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}

	/**
		Load a data item from the database using a statless session 
		when you know the id.
		@param session The Hibernate stateless session which is connected 
			to the database
		@param clazz An object of this class is returned with the data
		@param entityName The entity name of the item
		@param id The id of the item to load
	*/
	static <T> T loadData(StatelessSession session, Class<T> clazz, String entityName, int id)
		throws BaseException
	{
		assert session != null : "session == null";
		try
		{
			return clazz.cast(session.get(entityName, id));
		}
		catch(HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}

	
	/**
		Load a data item from the database using a query. If the query returns
		more than one item, only the first is returned.
		
		@param query A <code>Query</code> object which should select one
			item of the specified class
	*/
	static <T> T loadData(Query<T> query)
		throws BaseException
	{
		assert query != null : "query == null";
		try
		{
			List<T> l = query.list();
			return l.size() > 0 ? l.get(0) : null;
		}
		catch(HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}

	
	/**
		Exceute an update or delete query. 

		@param query A <code>Query</code> object which is executed using the
			{@link Query#executeUpdate()} method
		@return The number of affected items
	*/
	static int executeUpdate(Query<?> query)
		throws BaseException
	{
		assert query != null : "query == null";
		try
		{
			return query.executeUpdate();
		}
		catch(HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}

	/**
		Initialise a collection on a data object. This method is useful if you plan
		to detach an item.
		@see DbControl#detachItem(BasicItem)
	*/
	static void initCollection(Session session, BasicData data, String collectionName)
		throws BaseException
	{
		assert session != null : "session == null";
		assert data != null : "data == null";
		try
		{
			ClassMetadata cmd = getClassMetadata(Hibernate.getClass(data).getName());
			Object collection = cmd.getPropertyValue(data, collectionName);
			Hibernate.initialize(collection);
		}
		catch(HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}
	
	/**
		Delete a data item from the database. If the ID doesn't exist in the
		database this metod does nothing.
	*/
	static void deleteData(Session session, BasicData data)
		throws BaseException
	{
		assert session != null : "session == null";
		assert data != null : "data == null";
		try
		{
			session.delete(data);
		}
		catch(HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}
	
	/**
		Remove a data item from the internal cache of Hibernate. This also
		means that changes to the object will not be saved to the database
		unless it is reconnected again.
		@see #updateData(Session, BasicData)
		@see #lockData(Session, BasicData, LockOptions)
	*/
	static void evictData(Session session, BasicData data)
		throws BaseException
	{
		assert session != null : "session == null";
		if (data == null) return;
		try
		{
			session.evict(data);
		}
		catch(HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}
	
	/**
		Create a Hibernate query.
	*/
	@SuppressWarnings({ "unchecked", "rawtypes" })
	static <T> Query<T> createQuery(Session session, String hql)
		throws BaseException
	{
		assert session != null : "session == null";
		assert hql != null : "hql == null";
		try
		{
			Query<?> q = session.createQuery(hql);
			q.setFetchSize(getJdbcFetchSize());
			return new EntityQueryWrapper(q);
		}
		catch (HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}
	
	/**
		Create a Hibernate query with a known return type.
		@since 3.12
	*/
	static <R> Query<R> createQuery(Session session, String hql, Class<R> returnType)
		throws BaseException
	{
		assert session != null : "session == null";
		assert hql != null : "hql == null";
		try
		{
			Query<R> q = session.createQuery(hql, returnType);
			q.setFetchSize(getJdbcFetchSize());
			return new EntityQueryWrapper<R>(q);
		}
		catch (HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}

	/**
		Create a Hibernate query using a stateless session and a known return type.
		@since 3.12
	*/
	static <R> Query<R> createQuery(StatelessSession session, String hql, Class<R> returnType)
		throws BaseException
	{
		assert session != null : "session == null";
		assert hql != null : "hql == null";
		try
		{
			Query<R> q = session.createQuery(hql, returnType);
			q.setFetchSize(getJdbcFetchSize());
			return new EntityQueryWrapper<R>(q);
		}
		catch (HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}

	/**
		Create a Hibernate SQL query.
	*/
	static NativeQuery<?> createSqlQuery(Session session, String sql)
		throws BaseException
	{
		assert session != null : "session == null";
		assert sql != null : "sql == null";
		sql = sql.replace('[', getDialect().openQuote()).replace(']', getDialect().closeQuote());
		try
		{
			NativeQuery<?> q = session.createNativeQuery(sql);
			q.setFetchSize(getJdbcFetchSize());
			return new NativeQueryWrapper<>(q);
		}
		catch (HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}
	
	/**
		Create a Hibernate SQL query with a known return type.
		@since 3.12
	*/
	@SuppressWarnings("unchecked")
	static <R> NativeQuery<R> createSqlQuery(Session session, String sql, Class<R> returnType)
		throws BaseException
	{
		assert session != null : "session == null";
		assert sql != null : "sql == null";
		sql = sql.replace('[', getDialect().openQuote()).replace(']', getDialect().closeQuote());
		try
		{
			NativeQuery<R> q;
			// Specifying non-entity (for example, Integer) return type makes Hibernate go bananas 
			// Guess it is a bug, but until it has been fixed, we need to use the non-specific method call
			if (BasicData.class.isAssignableFrom(returnType))
			{
				q = session.createNativeQuery(sql, returnType);
			}
			else
			{
				q = session.createNativeQuery(sql);
			}
			q.setFetchSize(getJdbcFetchSize());
			return new NativeQueryWrapper<>(q);
		}
		catch (HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}

	/**
		Create a Hibernate SQL query using a stateless session.
	*/
	static NativeQuery<?> createSqlQuery(StatelessSession session, String sql)
		throws BaseException
	{
		assert session != null : "session == null";
		assert sql != null : "sql == null";
		sql = sql.replace('[', getDialect().openQuote()).replace(']', getDialect().closeQuote());
		try
		{
			NativeQuery<?> q = session.createNativeQuery(sql);
			q.setFetchSize(getJdbcFetchSize());
			return new NativeQueryWrapper<>(q);
		}
		catch (HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}
	
	/**
		Create a Hibernate SQL query with a known return type using a stateless session.
		@since 3.12
	*/
	@SuppressWarnings("unchecked")
	static <R> NativeQuery<R> createSqlQuery(StatelessSession session, String sql, Class<R> returnType)
		throws BaseException
	{
		assert session != null : "session == null";
		assert sql != null : "sql == null";
		sql = sql.replace('[', getDialect().openQuote()).replace(']', getDialect().closeQuote());
		try
		{
			NativeQuery<R> q;
			// Specifying non-entity (for example, Integer) return type makes Hibernate go bananas 
			// Guess it is a bug, but until it has been fixed, we need to use the non-specific method call
			if (BasicData.class.isAssignableFrom(returnType))
			{
				q = session.createNativeQuery(sql, returnType);
			}
			else
			{
				q = session.createNativeQuery(sql);
			}
			q.setFetchSize(getJdbcFetchSize());
			return new NativeQueryWrapper<>(q);
		}
		catch (HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}

	/**
		Enable the filter with the given name.
	*/
	static Filter enableFilter(Session session, String name)
		throws BaseException
	{
		assert session != null : "session == null";
		assert name != null : "name == null";
		try
		{
			return session.enableFilter(name);
		}
		catch (HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}
	
	/**
		Disable the filter with the given name.
	*/
	static void disableFilter(Session session, String name)
		throws BaseException
	{
		assert session != null : "session == null";
		assert name != null : "name == null";
		try
		{
			session.disableFilter(name);
		}
		catch (HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}
	
	/**
		Load a <code>List</code> of items from the database
		using a query. 
		
		@param query The query to execute
		@param sc A optional SessionControl that is automatically protected from
			timeouts during the time the query is running
	*/
	static <T> List<T> loadList(Query<T> query, SessionControl sc)
		throws BaseException
	{
		assert query != null : "query == null";
		Application.Pinger pinger = null;
		try
		{
			if (sc != null) pinger = Application.newPinger(sc);
			return query.list();
		}
		catch (Throwable ex)
		{
			throw new BaseException(ex);
		}
		finally
		{
			if (pinger != null) pinger.stop();
		}
	}

	/**
		Scroll through the result of a query using an iterator.
		@param query The query to execute
		@param sc A optional SessionControl that is automatically protected from
			timeouts during the time the query is running
		@since 3.12
	*/
	static <T> ScrollIterator<T> loadIterator(Query<T> query, boolean returnArray, SessionControl sc)
		throws BaseException
	{
		assert query != null : "query == null";
		Application.Pinger pinger = null;
		try
		{
			if (sc != null) pinger = Application.newPinger(sc);
			ScrollableResults result = query.scroll(ScrollMode.FORWARD_ONLY);
			return new ScrollIterator<T>(result, returnArray);
		}
		catch (Throwable ex)
		{
			throw new BaseException(ex);
		}
		finally
		{
			if (pinger != null) pinger.stop();
		}
	}
	
	/**
		Get a predefined HQL query.
		@see PredefinedQuery
	*/
	static Query<?> getPredefinedQuery(Session session, String name)
		throws BaseException
	{
		return getPredefinedQuery(session, name, (String[])null);
	}

	/**
		Get a predefined HQL query.
		@param replacements In the SQL string we look for {1}, {2}, ... and
			replace this with the contents of the string array
		@see PredefinedQuery
	*/
	static Query<?> getPredefinedQuery(Session session, String name, String... replacements)
		throws BaseException
	{
		assert session != null : "session == null";
		assert name != null : "name == null";
		return createQuery(session, PredefinedQuery.getQueryString(name, replacements));
	}

	/**
		Get a predefined HQL query with a known return type.
		@param replacements In the SQL string we look for {1}, {2}, ... and
			replace this with the contents of the string array
		@see PredefinedQuery
		@since 3.12
	*/
	static <R> Query<R> getPredefinedQuery(Session session, String name, Class<R> returnType, String... replacements)
		throws BaseException
	{
		assert session != null : "session == null";
		assert name != null : "name == null";
		return createQuery(session, PredefinedQuery.getQueryString(name, replacements), returnType);
	}

	/**
		Get a predefined HQL query with a known return type using the stateless session.
		@see PredefinedQuery
		@since 3.12
	*/
	static <R> Query<R> getPredefinedQuery(StatelessSession session, String name, Class<R> returnType)
		throws BaseException
	{
		assert session != null : "session == null";
		assert name != null : "name == null";
		return createQuery(session, PredefinedQuery.getQueryString(name), returnType);
	}

	/**
		Get a predefined SQL query.
		@see PredefinedQuery
	*/
	static Query<?> getPredefinedSQLQuery(Session session, String name, String... replacements)
		throws BaseException
	{
		assert session != null : "session == null";
		assert name != null : "name == null";
		return createSqlQuery(session, PredefinedQuery.getQueryString(name, replacements));
	}

	/**
		Get a predefined SQL query with a known return type.
		@see PredefinedQuery
		@since 3.12
	*/
	static <R> NativeQuery<R> getPredefinedSQLQuery(Session session, String name, Class<R> returnType, String... replacements)
		throws BaseException
	{
		assert session != null : "session == null";
		assert name != null : "name == null";
		return createSqlQuery(session, PredefinedQuery.getQueryString(name, replacements), returnType);
	}

	/**
		Get a predefined SQL query using the stateless session.
		@see PredefinedQuery
	*/
	static NativeQuery<?> getPredefinedSQLQuery(StatelessSession session, String name)
		throws BaseException
	{
		assert session != null : "session == null";
		assert name != null : "name == null";
		return createSqlQuery(session, PredefinedQuery.getQueryString(name));
	}
	
	/**
		Get a predefined SQL query with a known return type using the stateless session.
		@see PredefinedQuery
	*/
	static <R> NativeQuery<R> getPredefinedSQLQuery(StatelessSession session, String name, Class<R> returnType)
		throws BaseException
	{
		assert session != null : "session == null";
		assert name != null : "name == null";
		return createSqlQuery(session, PredefinedQuery.getQueryString(name), returnType);
	}

	static Iterator<PersistentClass> getClassMappings()
	{
		assert metadata != null : "HibernateUtil has not been initialised";
		return metadata.getEntityBindings().iterator();
	}

	static PersistentClass getClassMapping(String entityName)
	{
		assert metadata != null : "HibernateUtil has not been initialised";
		return metadata.getEntityBinding(entityName);
	}

	@SuppressWarnings("deprecation")
	static ClassMetadata getClassMetadata(String entityName)
	{
		assert sf != null : "HibernateUtil has not been initialised";
		return sf.getClassMetadata(entityName);
	}

	static Mapping getMapping()
	{
		return (Mapping)sf;
	}
	
	/**
		Get the database dialect.
	*/
	public static Dialect getDialect()
	{
		return dialect;
	}

	/**
		Get the database engine.
	*/
	public static DbEngine getDbEngine()
	{
		return dbEngine;
	}
	
	/**
		The number of rows to load at a time when reading 
		from the database.
		@since 3.9
	*/
	public static int getJdbcFetchSize()
	{
		return jdbcFetchSize;
	}
	
	/**
		Check if theta joins or ANSI joins are used by the database.
	 	@param queryType  Query type to be checked.
	 	@return TRUE if theta joins are used, FALSE if ANSI joins are used.
	*/
	public static boolean useThetaJoin(QueryType queryType)
	{
		return queryType == QueryType.SQL && dbEngine.useThetaJoin();
	}

	/**
		Quote a string with the default quote type for the current database engine.
		Ie. with MySQL: <code>value --&gt; `value`</code>, and with any ANSI-compatible
		database: <code>value --&gt; "value"</code>. This method uses the
		<code>Dialect.openQuote</code> and <code>Dialect.closeQuote</code> methods
		of the currently installed Hibernate dialect.
		
		@param value The value to quote
		@return The quoted value
		@see #getDialect()
	*/
	public static String quote(String value)
	{
		if (value != null) value = dialect.openQuote()+value+dialect.closeQuote();
		return value;
	}
	
	/**
		Removes the package name part of a class name and any ending 'Data'.
	*/
	static String getShortEntityName(String entityName)
	{
		int lastDot = entityName.lastIndexOf('.');
		if (lastDot >= 0)
		{
			entityName = entityName.substring(lastDot+1);
		}
		entityName = entityName.replaceAll("Data$", "");
		return entityName;
	}
	
	static void testTransactions()
		throws BaseException
	{
		NewsData news = new NewsData();
		news.setName("Transaction failure");
		news.setDescription("If you see this new item it is because your database doesn't support transactions. "+
			"We don't recommend running BASE without transactions.");
		news.setStartDate(new Date());
		news.setNewsDate(new Date());
		
		org.hibernate.Session session = null;
		org.hibernate.Transaction tx = null;
		boolean ok = false;
		try
		{
			// Insert news item, but rollback the transaction
			session = HibernateUtil.newSession();
			tx = HibernateUtil.newTransaction(session);
			HibernateUtil.saveData(session, news);
			int newsId = news.getId();
			HibernateUtil.rollback(tx);
			HibernateUtil.close(session);
			
			//  Evict the item from the cache and, in another session, try to load the news item
			sf.getCache().evictEntityData(NewsData.class, newsId);
			session = HibernateUtil.newSession();
			tx = HibernateUtil.newTransaction(session);
			news = HibernateUtil.loadData(session, NewsData.class, newsId);
			ok = news == null;
		}
		finally
		{
			if (tx != null) HibernateUtil.rollback(tx);
			if (session != null) HibernateUtil.close(session);
		}
		if (!ok) throw new BaseException("Your database doesn't support transactions.");
	}
	
	/**
		Prints a lot of useful stuff about the dynamic database to standard output.
		For each table the output includes information about columns, indexes,
		primary keys and foreign keys as found in the current database.
		
		@param verbose If true, lots of information will be printed
		@param silent If true, no information will be printed
		@param dropIndexes If true, all indexes will be dropped
	*/
	public static void dynamicDbIndexes(boolean verbose, final boolean silent, final boolean dropIndexes)
	{
		final boolean isVerbose = verbose && !silent;
		Session session = newSession();
		
		try
		{
			HibernateUtil.doWork(session, 
				new Work()
				{

					@Override
					public void execute(Connection connection) 
						throws SQLException
					{
						DatabaseMetaData metaData = null;
						List<Table> tables = new LinkedList<Table>();
						String dynamicCatalog = Application.getDynamicCatalog();
						String dynamicSchema = Application.getDynamicSchema();
						
						metaData = connection.getMetaData();
						ResultSet result = metaData.getTables(dynamicCatalog, dynamicSchema, null, null);
						
						while (result.next())
						{
							String tableName = result.getString("TABLE_NAME");
							String tableCatalog = result.getString("TABLE_CAT");
							String tableSchema = result.getString("TABLE_SCHEM");
							Table table = new Table();
							table.setName(tableName);
							table.setCatalog(tableCatalog);
							table.setSchema(tableSchema);
							tables.add(table);
						}
						
						for (Table table : tables)
						{
	
							if (!silent)
							{
								System.out.println("=================");
								System.out.println("Table   : " + table.getName());
							}
							if (isVerbose)
							{
								System.out.println("Catalog : " + table.getCatalog());
								System.out.println("Schema  : " + table.getSchema());
							}
	
							if (isVerbose)
							{
								System.out.println("Database information");
								System.out.println("--------------------");
							}
							
							TableInfo tiDb = null;
							try
							{
								tiDb = new TableInfo(table, metaData, dynamicCatalog, dynamicSchema);
							}
							catch (SQLException ex)
							{
								throw new BaseException(ex);
							}
	
							if (tiDb != null)
							{
								if (isVerbose)
								{
									// Write all columns
									for (ColumnInfo ci : tiDb.getColumns())
									{
										System.out.println("  Column       : " + ci);
									}
									
									// Write primary key
									System.out.println("  Primary key  : " + tiDb.getPrimaryKey());
								
									// Write foreign keys
									for (ForeignKeyInfo fk : tiDb.getForeignKeys())
									{
										System.out.println("  Foreign key  : " + fk);
									}
								}
							
								// Write indexes and unique constraints
								for (IndexInfo ii : tiDb.getIndexes())
								{
									if (!silent)
									{
										if (ii.isUnique())
										{
											System.out.println("  Unique       : " + ii);
										}
										else
										{
											System.out.println("  Index        : " + ii);
										}
									}
									
									boolean safeToDrop = tiDb.safeToDrop(ii);
	
									String dropSql = dbEngine.getDropIndexSql(
										table.getCatalog(), table.getSchema(), table.getName(), 
										ii.getName(), ii.isUnique());
									boolean actionDrop = dropIndexes && safeToDrop;
	
									if (!silent)
									{
										System.out.println("    Safe drop  : " + safeToDrop);
										System.out.println("    DROP-SQL   : " + dropSql);
										System.out.println("    Actions    : " + (actionDrop ? "DROP " : ""));
									}
									
									if (actionDrop )
									{
										Statement st = null;
										try
										{
											st = connection.createStatement();
											if (actionDrop) 
											{
												log.info("Dropping index: + " + dropSql);
												st.executeUpdate(dropSql);
											}
											connection.commit();
											st.close();
											st = null;
										}
										catch (SQLException ex)
										{
											log.error("Exception", ex);
											throw new BaseException(ex);
										}
										finally
										{
											if (st != null)
											{
												try
												{
													st.close();
												}
												catch (Throwable t)
												{
													log.error("Exception", t);
												}
											}
										}
									}
									
								}
							}
							
							if (!silent)
							{
								System.out.println("=================");
								System.out.println("");
							}
						}
					}
				}
			);
		}
		catch (SQLException ex)
		{
			throw new BaseException(ex);
		}
	}
	
	
	/**
		Prints a lot of useful stuff about the database to standard output.
		For each table the output includes information about columns, indexes,
		primary keys and foreign keys as found in the Hibernate mapping
		information and in the current database.
		
		@param verbose If true, lots of information will be printed
		@param silent If true, no information will be printed
		@param dropIndexes If true, all indexes will be dropped
		@param updateIndexes If true, indexes that doesn't already exist will be created
	*/
	public static void dbIndexes(boolean verbose, final boolean silent, 
		final boolean dropIndexes, final boolean updateIndexes)
	{
		Session session = newSession();
		try
		{
			HibernateUtil.doWork(session, new DbIndexWork(metadata, dialect, dbEngine, verbose, silent, dropIndexes, updateIndexes));
		}
		catch (SQLException ex)
		{}
	}

	/**
		Helper class for defining filters at runtime.
		@since 3.8
	*/
	static class FilterConfigurator
	{
		private final Map<String, Element> filterTemplates;
		private final List<DelayedFilterDef> delayedFilters;

		FilterConfigurator()
		{
			filterTemplates = new HashMap<String, Element>();
			delayedFilters = new ArrayList<DelayedFilterDef>();
		}
		
		void loadFilterTemplates(String source)
			throws IOException
		{
			URL filterTemplate = HibernateUtil.class.getResource(source);
			final String[] schemaFiles = {
				"http://www.hibernate.org/xsd/hibernate-mapping",
				HibernateUtil.class.getResource("/net/sf/basedb/core/xsd/hibernate-mapping-4.0.xsd").toString()
			};
			Document filterDef = XmlUtil2.getSchemaValidatedXML(filterTemplate, schemaFiles);
			for (Element filter : filterDef.getRootElement().getChildren())
			{
				filterTemplates.put(filter.getAttributeValue("name"), filter);
			}

		}
		
		/**
			Clone the &lt;filter-def&gt; with the given name and insert
			a new item-specific &lt;filter-def&gt; definition into the
			document given by the &lt;class&gt; element.
		*/
		void createItemSpecificFilterDef(String name, Item itemType, Element classTag)
		{
			// Clone the definition from the template and change the name
			Element filterDef = filterTemplates.get(name).clone();
			filterDef.setAttribute("name", name + itemType.name());
			// The <filter-def> should be a child to <hibernate-mapping>
			classTag.getDocument().getRootElement().addContent(filterDef);
		}
	
		/**
			Generic filters can be added as &lt;filter&lt; elements in the
			mapping file.
		*/
		void createGenericFilter(String name, String condition, Element classTag)
		{
			Element filter = new Element("filter", classTag.getNamespace());
			filter.setAttribute("name", name);
			filter.setAttribute("condition", condition);
			classTag.addContent(filter);
		}
		
		/**
			Item-specific filters are stored internally and registered later
			by {@link #registerDelayedFilters(org.hibernate.boot.Metadata)}.
			We need to do this since a &lt;filter&gt; tag is not allowed
			for items defined with &lt;subclass&gt;. Programmatically we can add
			the filter later but not until the configuration process has
			gone far enough so that we have access to PersistentClass instances.
		*/
		void createItemSpecificFilter(String name, Item itemType, String condition, Element classTag)
		{
			delayedFilters.add(new DelayedFilterDef(name, itemType, condition));
		}
		
		/**
			Register all filters that has been stored by
			{@link #createItemSpecificFilter(String, Item, String, Element)}
		*/
		void registerDelayedFilters(org.hibernate.boot.Metadata metadata)
		{
			for (DelayedFilterDef filter : delayedFilters)
			{
				PersistentClass pc = metadata.getEntityBinding(filter.itemType.getDataClass().getName());
				pc.addFilter(filter.name + filter.itemType.name(), filter.condition, true, null, null);
			}
		}
	}
		
	static class DelayedFilterDef
	{
		final String name;
		final Item itemType;
		final String condition;
		
		public DelayedFilterDef(String name, Item itemType, String condition) 
		{
			this.name = name;
			this.itemType = itemType;
			this.condition = condition;
		}
	}

}

