/**
	$Id: ItemSubtype.java 7381 2017-05-22 12:35:27Z nicklas $

	Copyright (C) 2011 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.lang.reflect.Method;
import java.util.Arrays;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import net.sf.basedb.core.data.ItemSubtypeData;
import net.sf.basedb.core.data.ItemSubtypeFileTypeData;
import net.sf.basedb.core.data.SubtypableData;
import net.sf.basedb.core.hibernate.TypeWrapper;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Orders;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.util.ClassUtil;

/**
	Item subtypes can be used to classify items into sub-categories.
	Subtypes can be defined for all item types that implement the
	{@link Subtypable} interface. For a given item type the name
	of the subtype should be unique.

	@author Nicklas
	@since 3.0
	@base.modified $Date: 2017-05-22 14:35:27 +0200 (Mon, 22 May 2017) $
*/
public class ItemSubtype
	extends BasicItem
	implements Nameable, Removable, Registered, SystemItem, Subtypable
{

	/**
		The type of item represented by this class.
		@see Item#ITEMSUBTYPE
		@see #getType()
	*/
	public static final Item TYPE = Item.ITEMSUBTYPE;

	
	private static Map<Item, List<Item>> relatedItems = new HashMap<Item, List<Item>>();
	/**
		Get a list with (subtypable) items that are related to the given
		item (which should also be subtypable). Typically, if the main item
		has a many-to-one relatation to another subtypable item, this item type
		will appear in the list. For example a {@link Protocol} has a link to
		{@link File} via {@link Protocol#getFile()}. Given the subtype of the protocol
		we can find the desired subtype of the file and check agains the actual subtype:
		
		<pre class="code">
Protocol p = ...
ItemSubtype protocolSubtype = p.getItemSubtype();
ItemSubtype desiredFileSubtype = protocolSubtype.getRelatedSubtype(Item.FILE);
File f = p.getFile();
ItemSubtype actualFileSubtype = f.getItemSubtype();
</pre>
		
		@return A list with the related items
		@see #getRelatedSubtype(Item)
	*/
	public static List<Item> getRelatedItems(Item mainItem)
	{
		if (!Subtypable.class.isAssignableFrom(mainItem.getItemClass())) return Collections.emptyList();

		List<Item> related = relatedItems.get(mainItem);
		if (related == null)
		{
			synchronized (relatedItems)
			{
				
				Method getItemSubtype = ClassUtil.findMethod(mainItem.getItemClass(), "getItemSubtype");
				SubtypableRelatedItems items = getItemSubtype == null ? null : getItemSubtype.getAnnotation(SubtypableRelatedItems.class);
				if (items == null)
				{
					related = Collections.emptyList();
				}
				else
				{
					related = Collections.unmodifiableList(Arrays.asList(items.value()));
				}
				relatedItems.put(mainItem, related);
			}
		}
		
		return related;
	}
	
	/**
		Check if the given item type has support for pushing annotations to
		the parent item or not.
		@since 3.1
	*/
	public static boolean supportsPushAnnotations(Item mainItem)
	{
		if (!Subtypable.class.isAssignableFrom(mainItem.getItemClass())) return false;

		Method getItemSubtype = ClassUtil.findMethod(mainItem.getItemClass(), "getItemSubtype");
		return getItemSubtype == null ? false : getItemSubtype.isAnnotationPresent(SupportsPushAnnotations.class);
	}
	
	/**
		Same as {@link #getRelatedSubtype(DbControl, Subtypable, Item, int, int)} with the
		same id for the last two parameters.
	*/
	public static ItemSubtype getRelatedSubtype(DbControl dc, Subtypable item, Item relatedItem, int defaultSubtype)
	{
		return getRelatedSubtype(dc, item, relatedItem, defaultSubtype, defaultSubtype);
	}
	
	/**
		Utility method for loading a subtype that is related to the subtype of the given item.
		This is more or less equivalent to calling {@link Subtypable#getItemSubtype()} and
		then {@link ItemSubtype#getRelatedSubtype(Item)} but with proper null handling and
		a possibility to return a default subtype if no related one can be found.
		
		@param dc A DbControl to use for database access
		@param item A subtypable item
		@param relatedItem The main item type of the related subtype we are looking for
		@param defaultIfNoSubtype The ID of the subtype to return if the item has no
			subtype
		@param defaultIfNoRelated The ID of the subtype to return if the item has a subtype,
			but the subtype has no related subtype for the given related item type
		@return An ItemSubtype or null
	*/
	public static ItemSubtype getRelatedSubtype(DbControl dc, Subtypable item, Item relatedItem, 
		int defaultIfNoSubtype, int defaultIfNoRelated)
	{
		ItemSubtype related = null;
		ItemSubtype subtype = item.getItemSubtype();
		if (subtype != null)
		{
			related = subtype.getRelatedSubtype(relatedItem);
			if (related == null && defaultIfNoRelated != 0) 
			{
				related = getById(dc, defaultIfNoRelated);
			}
		}
		else if (defaultIfNoSubtype != 0)
		{
			related = getById(dc, defaultIfNoSubtype);
		}
		return related;
	}
	
	/**
		Utility method for locating a subtype that is related to the subtype of
		a given item. This method will check each item type in turn and return
		the first subtype that is found.
		
		@param dc A DbControl to use for database access
		@param item A subtypable item
		@param relatedItems The main item type of the related subtype we are looking for
		@return An ItemSubtype or null
	*/
	public static ItemSubtype getRelatedSubtype(DbControl dc, Subtypable item, Item... relatedItems)
	{
		ItemSubtype subtype = item.getItemSubtype();
		if (subtype == null) return null;
		
		for (Item itemType : relatedItems)
		{
			ItemSubtype related = subtype.getRelatedSubtype(itemType);
			if (related != null) return related;
		}
		
		return null;
	}
	
	/**
		Utility method for locating all subtypes that have a relationship to the
		subtype of the given item. This is more or less equivalent to finding
		all subtypes were the method {@link ItemSubtype#getRelatedSubtype(Item)}
		return the subtype of the given subtypable item (optionally restricting
		to one parent item type).
		
		@param dc A DbControl to use for database access
		@param item A subtypeable item
		@param parentType The item type of the subtypes that we want to find or null
			to find any subtype
		@return A list with the subtypes, or an empty list if none could be found
	*/
	public static List<ItemSubtype> getParentSubtypes(DbControl dc, Subtypable item, Item parentType)
	{
		ItemSubtype subtype = item.getItemSubtype();
		if (subtype == null) return Collections.emptyList();
		
		ItemQuery<ItemSubtype> query = subtype.getParentSubtypes();
		if (parentType != null)
		{
			query.restrict(Restrictions.eq(Hql.property("itemType"), Expressions.integer(parentType.getValue())));
		}
		query.order(Orders.asc(Hql.property("name")));
		return query.list(dc);
	}
	
	/**
		Create a new item subtype item.
	
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
		@param itemType The main item type
		@return The new <code>ItemSubtype</code> item
		@throws BaseException If there is an error
	*/
	public static ItemSubtype getNew(DbControl dc, Item itemType)
		throws BaseException
	{
		ItemSubtype subtype = dc.newItem(ItemSubtype.class);
		subtype.setMainItemType(itemType);
		subtype.setName("New " + itemType.toString().toLowerCase() + " subtype");
		subtype.getData().setEntryDate(new Date());
		return subtype;
	}

	/**
		Get an item subtype object when you know the ID.
	
		@param dc The {@link DbControl} which will be used for
			permission checking and database access.
		@param id The system ID of the item to load.
		@return The <code>ItemSubtype</code> item
		@throws ItemNotFoundException This exception is thrown if an
			item with the specified ID is not found
		@throws PermissionDeniedException This exception is thrown if
			the logged in user doesn't have read permission to the item
		@throws BaseException This exception is thrown if there is another
			error
	*/
	public static ItemSubtype getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		ItemSubtype subtype = dc.loadItem(ItemSubtype.class, id);
		if (subtype == null) throw new ItemNotFoundException("ItemSubtype[ID="+id+"]");
		return subtype;
	}

	/**
		Get a query configured to retrieve item types.
		
		@param itemType Optionally filter the query to only return 
			subtypes for the given item type
		@return An {@link ItemQuery} object
	*/
	public static ItemQuery<ItemSubtype> getQuery(Item itemType)
	{
		ItemQuery<ItemSubtype> query = new ItemQuery<ItemSubtype>(ItemSubtype.class);
		if (itemType != null)
		{
			query.restrictPermanent(
				Restrictions.eq(Hql.property("itemType"), Expressions.integer(itemType.getValue()))
			);
		}
		return query;
	}

	ItemSubtype(ItemSubtypeData data)
	{
		super(data);
	}
	
	@Override
	ItemSubtypeData getData()
	{
		return (ItemSubtypeData)getBasicData();
	}
	
	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the Registered interface
		-------------------------------------------
	*/
	@Override
	public Date getEntryDate()
	{
		return DateUtil.copy(getData().getEntryDate());
	}
	@Override
	public void setEntryDate(Date entryDate)
	{
		if (isInDatabase()) 
		{
			throw new PermissionDeniedException("Not allowed to change entryDate of " + this);
		}
		getData().setEntryDate(DateUtil.setDateOrToday(entryDate));
	}
	// -------------------------------------------
	/*
		From the Removable interface
		-------------------------------------------
	*/
	@Override
	public boolean isRemoved()
	{
		return getData().getRemovedBy() != null;
	}
	@Override
	public void setRemoved(boolean removed)
		throws PermissionDeniedException
	{
		checkPermission(removed ? Permission.DELETE : Permission.WRITE);
		getData().setRemovedBy(removed ? getSessionControl().getLoggedInUserId() : null);
	}
	@Override
	public User getRemovedBy()
		throws PermissionDeniedException, ItemNotFoundException 
	{
		Integer userId = getData().getRemovedBy();
		return userId == null ? null : User.getById(getDbControl(), userId);
	}
	// -------------------------------------------
	/*
		From the SystemItem interface
		-------------------------------------------
	*/
	@Override
	public String getSystemId()
	{
		return getData().getSystemId();
	}
	
	@Override
	public boolean isSystemItem()
	{
		return getSystemId() != null;
	}
	// -------------------------------------------
	/*
		From the Nameable interface
		-------------------------------------------
	*/
	@Override
	public String getName()
	{
		return getData().getName();
	}
	
	@Override
	public void setName(String name)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setName(getData(), name);
	}
	
	@Override
	public String getDescription()
	{
		return getData().getDescription();
	}
	
	@Override
	public void setDescription(String description)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setDescription(getData(), description);
	}
	// -------------------------------------------
	/*
		From the Subtypable interface
		-------------------------------------------
	*/
	/**
		@return A reference to self
	*/
	@Override
	public ItemSubtype getItemSubtype()
	{
		return this;
	}
	/**
		Do nothing.
	*/
	@Override
	public void setItemSubtype(ItemSubtype subtype)
	{}
	// -------------------------------------------

	/*
		From the BasicItem class
		------------------------
	*/
	/**
		Checks if:
		<ul>
		<li>any {@link Subtypable} item is using this subtype.
		<li>For FILE subtypes: {@link MimeType#getFileType()} and {@link DataFileType#getGenericType()}.
		</ul>
	*/
	@Override
	public boolean isUsed()
		throws BaseException
	{
		Item mainType = getMainItemType();
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.query.Query<Long> query = HibernateUtil.getPredefinedQuery(session, 
				"GET_SUBTYPABLE_ITEMS_FOR_SUBTYPE_OF_CLASS", Long.class, "count(*)", mainType.getDataClass().getName());
			/*
				SELECT {1}
				FROM {2} sub
				WHERE sub.itemSubtype = :subtype
			*/
		query.setParameter("subtype", getId(), TypeWrapper.H_INTEGER);
		boolean used = HibernateUtil.loadData(query) > 0;
		if (!used)
		{
			// Item lists
			query = HibernateUtil.getPredefinedQuery(session, "GET_ITEMLISTS_FOR_SUBTYPE", Long.class, "count(*)");
			query.setParameter("subtype", getId(), TypeWrapper.H_INTEGER);
			used = HibernateUtil.loadData(query) > 0;
		}
		if (!used && mainType == Item.FILE)
		{
			query = HibernateUtil.getPredefinedQuery(session, "GET_MIMETYPES_FOR_FILETYPE", Long.class, "count(*)");
			query.setParameter("subtype", getId(), TypeWrapper.H_INTEGER);
			used = HibernateUtil.loadData(query) > 0;
		}
		if (!used && mainType == Item.FILE)
		{
			query = HibernateUtil.getPredefinedQuery(session, "GET_DATAFILETYPES_FOR_FILETYPE", Long.class, "count(*)");
			query.setParameter("subtype", getId(), TypeWrapper.H_INTEGER);
			used = HibernateUtil.loadData(query) > 0;
		}
		if (!used && mainType == Item.PROTOCOL)
		{
			query = HibernateUtil.getPredefinedQuery(session, "GET_PLATEEVENTTYPES_FOR_PROTOCOLTYPE", Long.class, "count(*)");
			query.setParameter("subtype", getId(), TypeWrapper.H_INTEGER);
			used = HibernateUtil.loadData(query) > 0;
		}
		return used || super.isUsed();
	}

	/**
		Get all:
		<ul>
		<li>{@link Subtypable} items of this subtype.
		</ul>
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		Item mainType = getMainItemType();
		org.hibernate.Session session = getDbControl().getHibernateSession();
	
		// Subtypable items
		org.hibernate.query.Query<Integer> query = HibernateUtil.getPredefinedQuery(session, 
				"GET_SUBTYPABLE_ITEMS_FOR_SUBTYPE_OF_CLASS", Integer.class, "sub.id", mainType.getDataClass().getName());
			/*
				SELECT {1}
				FROM {2} sub
				WHERE sub.itemSubtype = :subtype
			*/
		query.setParameter("subtype", getId(), TypeWrapper.H_INTEGER);
		addUsingItems(using, mainType, query);
		
		// Item lists
		query = HibernateUtil.getPredefinedQuery(session, "GET_ITEMLISTS_FOR_SUBTYPE", Integer.class, "il.id");
		query.setParameter("subtype", getId(), TypeWrapper.H_INTEGER);
		addUsingItems(using, Item.ITEMLIST, query);
		
		// File-related
		if (mainType == Item.FILE)
		{
			query = HibernateUtil.getPredefinedQuery(session, "GET_MIMETYPES_FOR_FILETYPE", Integer.class, "mt.id");
			query.setParameter("subtype", getId(), TypeWrapper.H_INTEGER);
			addUsingItems(using, Item.MIMETYPE, query);

			query = HibernateUtil.getPredefinedQuery(session, "GET_DATAFILETYPES_FOR_FILETYPE", Integer.class, "dft.id");
			query.setParameter("subtype", getId(), TypeWrapper.H_INTEGER);
			addUsingItems(using, Item.DATAFILETYPE, query);
		}
		// Protocol-related
		if (mainType == Item.PROTOCOL)
		{
			query = HibernateUtil.getPredefinedQuery(session, "GET_PLATEEVENTTYPES_FOR_PROTOCOLTYPE", Integer.class, "pet.id");
			query.setParameter("subtype", getId(), TypeWrapper.H_INTEGER);
			addUsingItems(using, Item.PLATEEVENTTYPE, query);
		}
		return using;
	}
	/**
		If this is a system protocol type, delete and create permissions are revoked.
	*/
	@Override
	void initPermissions(int granted, int denied)
		throws BaseException
	{
		if (isSystemItem()) 
		{
			denied |= Permission.deny(Permission.DELETE, Permission.CREATE);
		}
		super.initPermissions(granted, denied);
	}
	// -------------------------------------------
	
	/**
		Get the main item type that the subtype belongs to.
	*/
	public Item getMainItemType()
	{
		return Item.fromValue(getData().getItemType());
	}
	
	private void setMainItemType(Item itemType)
	{
		if (itemType == null) throw new InvalidUseOfNullException("itemType");
		getData().setItemType(itemType.getValue());
	}

	/**
		If this flag is set, the annotations on items of this subtype can be
		pushed to the parent item. Eg. the {@link Annotatable#getAnnotatableParents()}
		for an item may return items that are actually child items if they have a subtype with
		this flag set. For this feature to work it must also be supported by the
		actual parent item implementation of {@link Annotatable#getAnnotatableParents()}
		which is indicated by annotating the {@link Subtypable#getItemSubtype()} method
		with {@link SupportsPushAnnotations}.
		
		@return FALSE to use normal inheritance, TRUE to also inherit to parent items
		@since 3.1
	*/
	public boolean getPushAnnotations()
	{
		return getData().getPushAnnotations();
	}
	
	/**
		@see #getPushAnnotations()
	 	@since 3.1
	*/
	public void setPushAnnotations(boolean pushAnnotations)
	{
		checkPermission(Permission.WRITE);
		getData().setPushAnnotations(pushAnnotations);
	}
		
	
	/**
		Get a related subtype of another item type. For a given main type of 
		items use {@link #getRelatedItems(Item)} to find out which other 
		subtypable items that can be related to this subtype.
		
		@param itemType The other item type (which should also be subtypable)
		@return A subtype or null if no subtype has been configured.
	*/
	public ItemSubtype getRelatedSubtype(Item itemType)
	{
		return getDbControl().getItem(ItemSubtype.class, getData().getRelatedSubtypes().get(itemType.getValue()));
	}
	
	/**
		Set a subtype that is related to this subtype. Only one related subtype
		per item type is allowed. A subtype can't be related to itself.
		@param subtype The subtype (null is not allowed)
		@see #getRelatedSubtype(Item)
	*/
	public void setRelatedSubtype(ItemSubtype subtype)
	{
		checkPermission(Permission.WRITE);
		if (subtype == null) throw new InvalidUseOfNullException("subtype");
		if (this.equals(subtype))
		{
			throw new InvalidDataException("A subtype can't be related to itself: " + this.getName());
		}
		getData().getRelatedSubtypes().put(subtype.getMainItemType().getValue(), subtype.getData());
	}
	
	/**
		Remove the related subtype for the given item type.
		@param itemType The item type of the related subtype to
			be removed
	*/
	public void removeRelatedSubtype(Item itemType)
	{
		checkPermission(Permission.WRITE);
		getData().getRelatedSubtypes().remove(itemType.getValue());
	}
	
	/**
		Get a query returning all subtypes that are related to this subtype.
	*/
	public ItemQuery<ItemSubtype> getRelatedSubtypes()
	{
		ItemQuery<ItemSubtype> query = getQuery(null);
		query.joinPermanent(Hql.innerJoin("parents", "p"));
		query.restrictPermanent(Restrictions.eq(Hql.alias("p"), Hql.entity(this)));
		return query;
	}
	
	/**
		Get a query returning all subtypes that have a relation to this subtype.
	*/
	public ItemQuery<ItemSubtype> getParentSubtypes()
	{
		ItemQuery<ItemSubtype> query = getQuery(null);
		query.joinPermanent(Hql.innerJoin("relatedSubtypes", "p"));
		query.restrictPermanent(Restrictions.eq(Hql.alias("p"), Hql.entity(this)));
		return query;
	}
	
	/**
		Get a query returning the data file types that have been
		associated with this subtype. The association is only used
		for main item types that implemented the 
		{@link FileStoreEnabled} interface.
		@return A query
	*/
	public ItemQuery<ItemSubtypeFileType> getDataFileTypes()
	{
		ItemQuery<ItemSubtypeFileType> query = ItemSubtypeFileType.getQuery(this);
		return query;
	}
	
	/**
		Checks if the given data file type is associated with this subtype or not.
		@param fileType The file type to check
		@return TRUE if the file type is associated, FALSE if not
	*/
	public boolean isAssociatedDataFileType(DataFileType fileType)
	{
		return getData().getFileTypes().containsKey(fileType.getData());
	}
	
	/**
		Get information about a data file type that has been associated with this
		item subtype. The data file type must use the same main item type as this subtype.
		
		@param fileType The file type to get information for
		@param create If TRUE, the file type will be added to the subtype if it doesn't already exists
		@return Null if the file type isn't registered with the subtype (and create is false)
			or a {@link ItemSubtypeFileType} object
	*/
	public ItemSubtypeFileType getAssociatedDataFileType(DataFileType fileType, boolean create)
	{
		if (fileType == null) throw new InvalidUseOfNullException("fileType");
		
		ItemSubtypeFileType sf = null;
		ItemSubtypeFileTypeData sfData = getData().getFileTypes().get(fileType.getData());
		if (sfData != null)
		{
			sf = getDbControl().getItem(ItemSubtypeFileType.class, sfData);
		}
		else if (create)
		{
			checkPermission(Permission.WRITE);
			if (fileType.getItemType() != getMainItemType())
			{
				throw new InvalidDataException("The file type is not used for " + getMainItemType() + ": " + fileType.getName());
			}
			sf = ItemSubtypeFileType.getNew(getDbControl(), this, fileType);
			getData().getFileTypes().put(fileType.getData(), sf.getData());
			getDbControl().saveItemIf(this, sf, false);
		}
		return sf;
	}
	
	/**
		Remove the association between this subtype and the given data file type. 
		If the file type is not associated nothing happens.
		@param fileType The file type to remove
	*/
	public void removeAssociatedDataFileType(DataFileType fileType)
	{
		checkPermission(Permission.WRITE);
		if (getData().getFileTypes().containsKey(fileType.getData()))
		{
			ItemSubtypeFileTypeData sfData = getData().getFileTypes().remove(fileType.getData());
			DbControl dc = getDbControl();
			ItemSubtypeFileType sf = dc.getItem(ItemSubtypeFileType.class, sfData);
			dc.deleteItem(sf);
		}
	}
	
	/**
		Helper method for implementing the {@link Subtypable#setItemSubtype(ItemSubtype)}
		method.
		
		@param item The item which must be a subtypeable item
	*/
	public void setOnItem(Subtypable item)
	{
		if (item.getType() != this.getMainItemType())
		{
			throw new InvalidDataException("Subtype '" + getName() + "' can't be used on item " + item);
		}
		((SubtypableData)((BasicItem)item).getData()).setItemSubtype(getData());
	}
	
	/**
		Check if this subtype can be used on items of the given type.
		@param itemType The item type to check
	*/
	public void checkItemType(Item itemType)
	{
		if (itemType != getMainItemType())
		{
			throw new InvalidDataException("Subtype '" + getName() + "' can't be used on " + itemType);
		}
	}
	
}
