/*
	$Id: Plate.java 7381 2017-05-22 12:35:27Z nicklas $

	Copyright (C) 2005 Johan Enell, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.PlateData;
import net.sf.basedb.core.data.PlateEventData;
import net.sf.basedb.core.data.PlateCoordinate;
import net.sf.basedb.core.data.ReporterData;
import net.sf.basedb.core.data.WellData;
import net.sf.basedb.core.hibernate.TypeWrapper;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.PermissionRestriction;

import java.util.Date;
import java.util.Map;
import java.util.Set;
import java.util.HashSet;

/**
	This class is represents physical plates. A plate must have
	a {@link PlateType}, which defines the geometry and possible events
	for the plate.
	<p>
	A plate consists of {@link Well}:s which are created at the same
	time the plate is created. The wells are organised into rows and
	columns as specified by the plate geometry.
	<p>
	A plate is either created directly or as the result of a {@link
	PlateMapping} operation from other plates. In the first case the
	wells must be linked to reporters before the plate is saved. In
	the latter case the linking is done automatically from the
	source plates.

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2017-05-22 14:35:27 +0200 (Mon, 22 May 2017) $
*/
public class Plate
	extends AnnotatedItem
	implements Registered
{
	/**
		The type of item represented by this class.
		@see Item#PLATE
		@see #getType()
	*/
	public static final Item TYPE = Item.PLATE;
	
	/**
		Create a new <code>Plate</code> item.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
		@param plateType The type of the new plate
		@return The new <code>Plate</code> item
		@throws PermissionDeniedException If the loggged in user
			doesn't have use permission for the plate type
		@throws InvalidDataException If the plate type is null
		@throws BaseException If there is another error
	*/
	public static Plate getNew(DbControl dc, PlateType plateType)
		throws BaseException
	{
		Plate p = dc.newItem(Plate.class);
		p.setName("New plate");
		p.setPlateType(plateType);
		int rows = plateType.getData().getPlateGeometry().getRows();
		int columns = plateType.getData().getPlateGeometry().getColumns();
		PlateData plateData = p.getData();
		plateData.setEntryDate(new Date());
		Map<PlateCoordinate, WellData> wells = plateData.getWells();
		for (int row = 0; row < rows; ++row)
		{
			for (int column = 0; column < columns; ++column)
			{
				WellData well = new WellData();
				well.setPlate(plateData);
				well.setRow(row);
				well.setColumn(column);
				wells.put(new PlateCoordinate(row, column), well);
			}
		}
		return p;
	}

	/**
		Get a <code>Plate</code> item when you know the id.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The id of the item to load
		@return The <code>Plate</code> item
		@throws ItemNotFoundException If an item with the specified 
			id is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static Plate getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		Plate p = dc.loadItem(Plate.class, id);
		if (p == null) throw new ItemNotFoundException("Plate[id="+id+"]");
		return p;
	}

	/**
		Get a query that returns plates.
		@return An {@link ItemQuery} object
	*/
	public static ItemQuery<Plate> getQuery()
	{
		return new ItemQuery<Plate>(Plate.class);
	}

	Plate(PlateData plateData)
	{
		super(plateData);
	}

	@Override
	PlateData getData()
	{
		return (PlateData)getBasicData();
	}
	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/**
		Get the parent plates, if any.
	*/
	@Override
	public Set<Annotatable> getAnnotatableParents()
		throws BaseException
	{
		Set<Annotatable> annotatable = new HashSet<Annotatable>();
		ItemQuery<Plate> plates = getParentPlates();
		plates.include(Include.ALL);
		annotatable.addAll(plates.list(getDbControl()));
		return annotatable;
	}
	// -------------------------------------------
	/*
		From the Registered interface
		-------------------------------------------
	*/
	@Override
	public Date getEntryDate()
	{
		return DateUtil.copy(getData().getEntryDate());
	}
	@Override
	public void setEntryDate(Date entryDate)
	{
		if (isInDatabase()) 
		{
			throw new PermissionDeniedException("Not allowed to change entryDate of " + this);
		}
		getData().setEntryDate(DateUtil.setDateOrToday(entryDate));
	}
	// -------------------------------------------
	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		Check if:
		<ul>
		<li>another <code>Plate</code> has been created from this one (via a {@link PlateMapping})
		<li>any {@link ArrayDesign}:s are using this plate
		</ul>
		Wells are deleted automatically.
	*/
	@Override
	public boolean isUsed()
		throws BaseException
	{
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.query.Query<Long> query = HibernateUtil.getPredefinedQuery(session, 
			"GET_CHILDPLATES_FOR_PLATE", Long.class, "count(*)");
		/*
			SELECT {1}
			FROM PlateData p
			JOIN p.parents pp
			WHERE pp = :plate
		*/
		query.setParameter("plate", getId(), TypeWrapper.H_INTEGER);
		boolean used = HibernateUtil.loadData(query) > 0;
		if (!used)
		{
			query = HibernateUtil.getPredefinedQuery(session, 
				"GET_ARRAYDESIGNS_FOR_PLATE", Long.class, "count(*)");
			/*
				SELECT {1}
				FROM ArrayDesignData ad
				JOIN ad.plates p
				WHERE p = :plate
			*/
			query.setParameter("plate", getId(), TypeWrapper.H_INTEGER);
			used = HibernateUtil.loadData(query) > 0;
		}
		return used || super.isUsed();
	}
	/**
		Get all:
		<ul>
		<li>Child plates
		<li>{@link ArrayDesign}:s using this plate 
		</ul>
		@since 2.2
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		org.hibernate.Session session = getDbControl().getHibernateSession();
		
		// Plates
		org.hibernate.query.Query<Integer> query = HibernateUtil.getPredefinedQuery(session, 
				"GET_CHILDPLATES_FOR_PLATE", Integer.class, "p.id");
			/*
				SELECT {1}
				FROM PlateData p
				JOIN p.parents pp
				WHERE pp = :plate
			*/
		query.setParameter("plate", getId(), TypeWrapper.H_INTEGER);
		addUsingItems(using, Item.PLATE, query);
		
		// Array designs
		query = HibernateUtil.getPredefinedQuery(session, 
				"GET_ARRAYDESIGNS_FOR_PLATE", Integer.class, "ad.id");
			/*
				SELECT {1}
				FROM ArrayDesignData ad
				JOIN ad.plates p
				WHERE p = :plate
			*/
		query.setParameter("plate", getId(), TypeWrapper.H_INTEGER);
		addUsingItems(using, Item.ARRAYDESIGN, query);
		
		return using;
	}
	// -------------------------------------------

	/**
		Get the {@link PlateType} of this plate.
		@return The <code>PlateType</code> item
		@throws PermissionDeniedException If the logged in user doesn't have
			{@link Permission#READ} permission for the plate type
		@throws BaseException If there is another error
	*/
	public PlateType getPlateType()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(PlateType.class, getData().getPlateType()); 
	}
	
	/**
		Set the plate type of this plate.
		@throws PermissionDeniedException If the logged in user
			doesn't have use permission for the plate type
		@throws InvalidDataException If the plate type is null
	*/
	private void setPlateType(PlateType plateType)
		throws PermissionDeniedException, InvalidDataException
	{
		if (plateType == null) throw new InvalidUseOfNullException("plateType");
		plateType.checkPermission(Permission.USE);
		getData().setPlateType(plateType.getData());
	}
	
	/**
		Get the {@link PlateMapping} used to create this plate.
		@return The <code>PlateMapping</code> item, or null if this plate was not
			created using a mapping
		@throws PermissionDeniedException If the logged in user doesn't have
			read permission to the plate mapping
		@throws BaseException If there is another error
	*/
	public PlateMapping getPlateMapping()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(PlateMapping.class, getData().getPlateMapping()); 
	}

	/**
		If the plate was created by a plate mapping, this is the index
		of this plate in the list of created plates. First plate have index 0.
	*/
	public int getDestinationIndex()
	{
		return getData().getDestinationIndex();
	}
	
	/**
		The maximum length of the barcode that can be stored in the
		database. Check the length against this value before calling the
		{@link #setBarcode(String)} method to avoid exceptions.
	*/
	public static final int MAX_BARCODE_LENGTH = PlateData.MAX_BARCODE_LENGTH;
	/**
		Get the barcode of this plate. The barcode is not used by the 
		core and it is not required that it is unique.
	*/
	public String getBarcode()
	{
		return getData().getBarcode();
	}
	/**
		Set the barcode for the plate. Null is allowed but the value 
		can't be longer than the value specified by
		the {@link #MAX_BARCODE_LENGTH} constant. Normally,
		this should be a unique value among all plates, but the core does not
		check this.
		@param barcode The new barcode for the plate
		@throws PermissionDeniedException If the logged in user doesn't
			have write permission
		@throws InvalidDataException If the value is too long
	*/
	public void setBarcode(String barcode)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		getData().setBarcode(StringUtil.setNullableString(barcode, "barcode", MAX_BARCODE_LENGTH));
	}
	
	/**
		Check if the plate has been destroyed. A destroyed plate
		can't be used by a {@link PlateMapping} to create new plates
		or to create {@link ArrayDesign}:s.
	*/
	public boolean isDestroyed()
	{
		return getData().isDestroyed();
	}
	
	/**
		Set the destroyed flag. A destroyed plate
		can't be used by a {@link PlateMapping} to create new plates
		or to create {@link ArrayDesign}:s.
	 	@param destroyed TRUE if the plate is destroyed, FALSE otherwise. 
	 	@throws PermissionDeniedException If the logged in user doesn't 
	 		have write permission on the item.
	*/
	public void setDestroyed(boolean destroyed)
		throws PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		getData().setDestroyed(destroyed);
	}
	
	/**
		Get a query returning the parent plates of this plate. A plate
		has parents if it was created by a {@link PlateMapping}.
		@return An <code>ItemQuery</code> object
	*/
	public ItemQuery<Plate> getParentPlates()
	{
		ItemQuery<Plate> query = Plate.getQuery();
		query.joinPermanent(Hql.innerJoin("children", "c"));
		query.restrictPermanent(
			Restrictions.eq(
				Hql.alias("c"), 
				Hql.entity(this)
			)
		);
		return query;
	}
	
	/**
		If this plate was created by a {@link PlateMapping} and the specified plate
		is a parent, get the index of the parent plate in the list
		of source plates used by the plate mapping. If the plate is not a parent, -1
		is returned.
	 	@param parent The plate to get source index of, if it's a parent. Null is not allowed. 
		@return The source index or -1 if the plate is not a parent plate
		@throws InvalidDataException If the parent is null
	*/
	public int getSourceIndex(Plate parent)
		throws InvalidDataException
	{
		if (parent == null) throw new InvalidDataException("parent");
		return IntegerUtil.getInt(getData().getParents().get(parent.getData()), -1);
	}
	
	/**
		Get a query returning the child plates of this plate. 
		@return An <code>ItemQuery</code> object
	*/
	public ItemQuery<Plate> getChildPlates()
	{
		ItemQuery<Plate> query = Plate.getQuery();
		query.joinPermanent(Hql.innerJoin("parents", "p"));
		query.restrictPermanent(
			Restrictions.eq(
				Hql.index("p", null), 
				Hql.entity(this)
			)
		);
		return query;
	}
	
	/**
		Get the event for the specified event type. If no
		event for that type exist a new one is created.
		@param eventType The event type to get or create the event for
		@return The <code>PlateEvent</code> item
		@throws PermissionDeniedException If the logged in user
			doesn't have read permission for an existing event
			or doesn't have write permission if a new event must be 
			created
		@throws InvalidDataException If the event type is null or
			doesn't belong to the same {@link PlateType} as the plate
		@throws BaseException If there is another error
		@see #hasEvent(PlateEventType)
		@see #getEvents()
	*/
	public PlateEvent getEvent(PlateEventType eventType)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		if (eventType == null) throw new InvalidUseOfNullException("eventType");
		PlateEventData eventData = getData().getEvents().get(eventType.getData());
		PlateEvent event = null;
		if (eventData == null)
		{
			event = PlateEvent.getNew(getDbControl(), this, eventType);
		}
		else
		{
			event = getDbControl().getItem(PlateEvent.class, eventData);
		}
		return event;
	}
	
	/**
		Check if an event for the specified event type has been created
		for this plate. A plate may only have one event of each type.
		@param eventType The event type to check
		@return TRUE if an event of the specified type exists, FALSE otherwise
		@see #getEvent(PlateEventType)
		@see #getEvents()
	*/
	public boolean hasEvent(PlateEventType eventType)
	{
		if (eventType == null) return false;
		return getData().getEvents().containsKey(eventType.getData());
	}
	

	/**
		Get a query that returns all event types for this plate type.
		@return An {@link ItemQuery} object
		@throws BaseException If there is an error
	*/
	public ItemQuery<PlateEvent> getEvents()
	{
		return PlateEvent.getQuery(this);
	}
	
	/**
		Get the well on the specified plate coordinate.
		@param row The row coordinate of the well. Starting at 0.
		@param column The column coordinate of the well. Starting at 0.
		@return A <code>Well</code> item or null if the coordinates
			are outside the range specified by the geometry
		@throws PermissionDeniedException If the logged in user doesn't
			have read permission for the well
		@throws BaseException If there is another error
	*/
	public Well getWell(int row, int column)
		throws PermissionDeniedException, BaseException
	{
		WellData wellData = getData().getWells().get(new PlateCoordinate(row, column));
		return getDbControl().getItem(Well.class, wellData);
	}
	
	/**
		Get a query returning all wells for this plate.
		@return An <code>ItemQuery</code> object
		@throws BaseException If there is an error
	*/
	public ItemQuery<Well> getWells()
	{
		return Well.getQuery(this);
	}
	
	/**
		Get a query that returns the reporters on this plate. The query
		is rooted at the wells, which means that:
		<ul>
		<li>All filters, sorting, etc. that works for {@link #getWells()} can be 
			used without modifications on this query.
		<li>The same reporter may be returned multiple times if it is present on several
			wells.
		<li>'Null' values may be returned if there are wells without a reporter
		</ul>
		
		@return A {@link DataQuery} object
		@since 2.8
	*/
	public DataQuery<ReporterData> getReporters()
	{
		DataQuery<ReporterData> query = 
			new DataQuery<ReporterData>(ReporterData.class, Item.WELL, null, "reporter");
		// We allow reporters without READ permission unless it has been explicitely DENIED
		query.restrictPermanent(new PermissionRestriction(Permission.DENIED, Item.REPORTER));
		query.restrictPermanent(
				Restrictions.eq(
					Hql.property("plate"),
					Hql.entity(this)
				)
			);
		return query;
	}
	
}
