/*
	$Id: PlateEvent.java 7308 2017-03-13 08:15:46Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.PlateEventData;
import net.sf.basedb.core.data.SharedData;
import net.sf.basedb.core.data.UserData;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Hql;

import java.util.Date;

/**
	This class represents an event in the life of a {@link Plate}.
	The possible events for a plate are defined by the 
	{@link PlateEventType}:s of a the plates {@link PlateType}.

	@author Nicklas
	@version 2.0
	@see Plate
	@see PlateEventType
	@base.modified $Date: 2017-03-13 09:15:46 +0100 (Mon, 13 Mar 2017) $
*/
public class PlateEvent
	extends ChildItem
	implements Registered
{
	/**
		The type of item represented by this class.
		@see Item#PLATEEVENT
		@see #getType()
	*/
	public static final Item TYPE = Item.PLATEEVENT;
	
	/**
		This filter will only return items if the logged in user has
		generic read permission to plates.
	*/
	private static final QueryRuntimeFilter RUNTIME_FILTER = 
		new QueryRuntimeFilterFactory.ChildFilter(TYPE, Item.PLATE);

	/**
		Create a new <code>PlateEvent</code> item.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
		@param plate The plate
		@param plateEventType The type of the event
		@return The new <code>PlateEvent</code> item
		@throws PermissionDeniedException If the loggged in user
			doesn't have write permission for the plate type
		@throws InvalidDataException If the plate or event type is null,
			or if the plate type of the plate and event type are different
		@throws BaseException If there is an error
	*/
	public static PlateEvent getNew(DbControl dc, Plate plate, PlateEventType plateEventType)
		throws InvalidDataException, BaseException
	{
		PlateEvent pe = dc.newItem(PlateEvent.class);
		pe.setPlate(plate);
		pe.setPlateEventType(plateEventType);
		pe.getData().setEntryDate(new Date());
		org.hibernate.Session session = dc.getHibernateSession();
		SessionControl sc = dc.getSessionControl();
		pe.getData().setUser(HibernateUtil.loadData(session, UserData.class, sc.getLoggedInUserId()));
		return pe;
	}

	/**
		Get a <code>PlateEvent</code> item when you know the id.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The id of the item to load
		@return The <code>PlateEvent</code> item
		@throws ItemNotFoundException If an item with the specified 
			id is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static PlateEvent getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		PlateEvent pe = dc.loadItem(PlateEvent.class, id);
		if (pe == null) throw new ItemNotFoundException("PlateEvent[id="+id+"]");
		return pe;
	}

	/**
		Get a query configured to retrieve plate events for the specified plate.
		
		@param plate The plate to retreive event for, null is allowed if
			the logged in user has generic READ permission for plates in which case
			all events will be returned
		@return An {@link ItemQuery} object
		@see Plate#getEvents()
	*/
	public static ItemQuery<PlateEvent> getQuery(Plate plate)
	{
		ItemQuery<PlateEvent> query = null;
		if (plate != null)
		{
			query = new ItemQuery<PlateEvent>(PlateEvent.class, null);
			query.restrictPermanent(
				Restrictions.eq(
					Hql.property("plate"), 
					Hql.entity(plate)
				)
			);
		}
		else
		{
			query = new ItemQuery<PlateEvent>(PlateEvent.class, RUNTIME_FILTER);
		}
		return query;
	}

	PlateEvent(PlateEventData plateEventData)
	{
		super(plateEventData);
	}

	@Override
	PlateEventData getData()
	{
		return (PlateEventData)getBasicData();
	}
	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the ChildItem class
		-------------------------------------------
	*/
	@Override
	SharedData getSharedParent()
	{
		return getData().getPlate();
	}
	// -------------------------------------------

	/**
		Get the {@link Plate} this event belongs to.
		@return The <code>Plate</code> item
		@throws PermissionDeniedException If the logged in user doesn't have
			{@link Permission#READ} permission for the plate
		@throws BaseException If there is another error
	*/
	public Plate getPlate()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(Plate.class, getData().getPlate());
	}
	/**
		Set the plate owning this event.
		@throws PermissionDeniedException If the logged in user
			doesn't have write permission for the plate
		@throws InvalidDataException If the plate is null
	*/
	private void setPlate(Plate plate)
		throws PermissionDeniedException, InvalidDataException
	{
		if (plate == null) throw new InvalidUseOfNullException("plate");
		plate.checkPermission(Permission.WRITE);
		getData().setPlate(plate.getData());
	}

	/**
		Get the {@link PlateEventType} of this event.
		@return The <code>PlateEventType</code> item
		@throws PermissionDeniedException If the logged in user doesn't have
			{@link Permission#READ} permission for the event type
		@throws BaseException If there is another error
	*/
	public PlateEventType getPlateEventType()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(PlateEventType.class, getData().getPlateEventType());
	}
	/**
		Set the plate event type of this event.
		@throws InvalidDataException If the event type is null
			or belongs to a different {@link PlateType} than the
			plate
	*/
	private void setPlateEventType(PlateEventType plateEventType)
		throws PermissionDeniedException, InvalidDataException
	{
		if (plateEventType == null) throw new InvalidUseOfNullException("plateEventType");
		if (!plateEventType.getData().getPlateType().equals(getData().getPlate().getPlateType()))
		{
			throw new InvalidDataException("The plate type of the event and plate does not match");
		}
		getData().setPlateEventType(plateEventType.getData());
	}

	/**
		Get the {@link Protocol} that was followed in this event.
		@return The <code>Protocol</code> item, or null
			if no protocol type has been specified
		@throws PermissionDeniedException If the logged in user doesn't
			have read permission to the protocol 
		@throws BaseException If there is another error
	*/
	public Protocol getProtocol()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(Protocol.class, getData().getProtocol());
	}

	/**
		Set the protocol that was followed in this event. It is recommended
		that the {@link ItemSubtype} of the protocol matches the 
		protocol type set by the {@link PlateEventType}, but the core doesn't
		check this.

		@param protocol The new <code>Protocol</code>
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission for this event or use permission for the
			protocol
		@throws BaseException If there is another error
	*/
	public void setProtocol(Protocol protocol)
		throws PermissionDeniedException, BaseException
	{
		checkPermission(Permission.WRITE);
		if (protocol != null) protocol.checkPermission(Permission.USE);
		getData().setProtocol(protocol == null ? null : protocol.getData());
	}
	
	/**
	 	Get the {@link Hardware} that was used in this event.
	 	@return A <code>Hardware</code>item or null of not known
	  	@throws PermissionDeniedException If the logged in user doesn't have
	  		read permission to the hardware.	 
	  	@throws BaseException If there is another error.
	*/
	public Hardware getHardware()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(Hardware.class, getData().getHardware());
	}
	
	/**
	 	Set the {@link Hardware} that was used in this event.
	 	@param hardware A <code>Hardware</code> item or null if not known.
	 	@throws PermissionDeniedException If the logged in user doesn't have 
	 		write permission on this event or use permission for the hardware.
	*/
	public void setHardware(Hardware hardware)
		throws PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		if (hardware != null) hardware.checkPermission(Permission.USE);
		getData().setHardware(hardware == null ? null : hardware.getData());
	}

	/**
		The maximum length of the comment about this event. Check the length
		against this value before calling the {@link #setComment(String)}
		method to avoid exceptions.
	*/
	public static final int MAX_COMMENT_LENGTH = PlateEventData.MAX_COMMENT_LENGTH;
	/**
		Get the comment about this event.
		@return The comment
	*/
	public String getComment()
	{
		return getData().getComment();
	}
	/**
		Set the comment about this event. The value may be null, but mustn't
		be longer than the value specified by the {@link #MAX_COMMENT_LENGTH} constant.
		@param comment The new comment
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission on this event
		@throws InvalidDataException If the comment is too long
	*/
	public void setComment(String comment)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		getData().setComment(StringUtil.setNullableString(comment, "comment", MAX_COMMENT_LENGTH));
	}
	
	/**
		Get the date this event was added to the database.
		@return A <code>Date</code> object
	*/
	@Override
	public Date getEntryDate()
	{
		return DateUtil.copy(getData().getEntryDate());
	}
	@Override
	public void setEntryDate(Date entryDate)
	{
		if (isInDatabase()) 
		{
			throw new PermissionDeniedException("Not allowed to change entryDate of " + this);
		}
		getData().setEntryDate(DateUtil.setDateOrToday(entryDate));
	}

	/**
		Get the date this event was done in the lab.
		@return A <code>Date</code> object, or null if not known
	*/
	public Date getEventDate()
	{
		return DateUtil.copy(getData().getEventDate());
	}
	/**
		Set the date this event was done in the lab.
		@param eventDate The date, or null if not known
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission on this event
	*/
	public void setEventDate(Date eventDate)
		throws PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		getData().setEventDate(DateUtil.setNullableDate(eventDate, "eventDate"));
	}

	/**
		Get the {@link User} that is responsible for this event.
		@return A <code>User</code> item 
		@throws PermissionDeniedException If the logged in user doesn't have
			read permission to the user
		@throws BaseException If there is another error
	*/
	public User getUser()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(User.class, getData().getUser());
	}


}
