/**
	$Id$

	Copyright (C) 2011 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.migrate.postgres;

import java.io.Writer;
import java.sql.ResultSet;
import java.sql.Types;

import org.apache.commons.lang3.time.FastDateFormat;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.dbengine.TableInfo.ColumnInfo;
import net.sf.basedb.core.migrate.ByteArrayWriter;
import net.sf.basedb.core.migrate.BooleanWriter;
import net.sf.basedb.core.migrate.DateWriter;
import net.sf.basedb.core.migrate.DoubleWriter;
import net.sf.basedb.core.migrate.FloatWriter;
import net.sf.basedb.core.migrate.IntegerWriter;
import net.sf.basedb.core.migrate.LongWriter;
import net.sf.basedb.core.migrate.ResultWriter;
import net.sf.basedb.core.migrate.ResultWriterFactory;
import net.sf.basedb.core.migrate.ShortWriter;
import net.sf.basedb.core.migrate.TextWriter;
import net.sf.basedb.core.migrate.TimestampWriter;

/**
 	Result writer factory implementation for formatting
 	result values suitable for importing by a PostgreSQL
 	database.
	
	@author nicklas
	@since 3.1
 */
public class PostgresResultWriterFactory 
	implements ResultWriterFactory
{

	/**
		PostgreSQL uses "\N" for null values.
	*/
	private final char[] NULL_ESCAPE;
	/**
		PostgreSQL uses 'yyyy-MM-dd' for dates.
	 */
	private final FastDateFormat dateFormat; 
	/**
		PostgreSQL uses 'yyyy-MM-dd HH:mm:ss' for timestamps.
	 */
	private final FastDateFormat timestampFormat; 
	
	private char[][] BINARY_ESCAPE;
	
	private char[][] CHAR_ESCAPE;
	
	public PostgresResultWriterFactory()
	{
		this.NULL_ESCAPE = new char[] { '\\', 'N' };
		this.dateFormat = FastDateFormat.getInstance("yyyy-MM-dd");
		this.timestampFormat = FastDateFormat.getInstance("yyyy-MM-dd HH:mm:ss");
	}
	
	@Override
	public ResultWriter create(ResultSet rs, int index, Writer writer, ColumnInfo column)
	{
		ResultWriter rw;
		int sqlType = column.getSqlType();
		switch (sqlType)
		{
			case Types.VARCHAR:
			case Types.LONGVARCHAR:
			{
				if (CHAR_ESCAPE == null)
				{
					CHAR_ESCAPE = new char[100][];
					CHAR_ESCAPE[0] = new char[0]; // 0-byte is ignored in strings
					CHAR_ESCAPE['\n'] = new char[] { '\\', 'n' };
					CHAR_ESCAPE['\r'] = new char[] { '\\', 'r' };
					CHAR_ESCAPE['\t'] = new char[] { '\\', 't' };
					CHAR_ESCAPE['\\'] = new char[] { '\\', '\\' };
				}
				rw = new TextWriter(rs, index, writer, NULL_ESCAPE, CHAR_ESCAPE);
				break;
			}
			case Types.SMALLINT:
			{
				rw = new ShortWriter(rs, index, writer, NULL_ESCAPE);
				break;
			}
			case Types.BIGINT:
			{
				rw = new LongWriter(rs, index, writer, NULL_ESCAPE);
				break;
			}
			case Types.INTEGER:
			{
				rw = new IntegerWriter(rs, index, writer, NULL_ESCAPE);
				break;
			}
			case Types.FLOAT:
			case Types.REAL:
			{
				rw = new FloatWriter(rs, index, writer, NULL_ESCAPE);
				break;
			}
			case Types.DOUBLE:
			{
				rw = new DoubleWriter(rs, index, writer, NULL_ESCAPE);
				break;
			}
			case Types.DATE:
			{
				rw = new DateWriter(rs, index, writer, NULL_ESCAPE, dateFormat);
				break;
			}
			case Types.TIMESTAMP:
			{
				rw = new TimestampWriter(rs, index, writer, NULL_ESCAPE, timestampFormat);
				break;
			}
			case Types.BIT:
			case Types.BOOLEAN:
			{
				rw = new BooleanWriter(rs, index, writer, NULL_ESCAPE);
				break;
			}
			case Types.BINARY:
			case Types.VARBINARY:
			case Types.LONGVARBINARY:
			{
				if (BINARY_ESCAPE == null)
				{
					// Setup the binary escape matrix
					// Values below 32 and above 127 need to be escape with a \\+octal representation
					// eg. \000
					// A special case is 92 == '\' which must also be escaped with '\\\\'
					BINARY_ESCAPE = new char[256][];
					for (int byteValue = 0; byteValue < 256; ++byteValue)
					{
						if (byteValue < 32 || byteValue > 127)
						{
							// Escape with \ + octal value (zero-padded to 3 digits)
							char[] escape = new char[5];
							escape[0] = '\\';
							escape[1] = '\\';
							String octal = Integer.toOctalString(byteValue);
							int pad = octal.length()-3;
							for (int i = 2; i < 5; ++i, ++pad)
							{
								escape[i] = pad < 0 ? '0' : octal.charAt(pad);
							}
							BINARY_ESCAPE[byteValue] = escape;
						}
						else if (byteValue == 92) // 92 == '\'
						{
							BINARY_ESCAPE[byteValue] = new char[] { '\\', '\\', '\\', '\\' };
						}
					}
				}
				rw = new ByteArrayWriter(rs, index, writer, NULL_ESCAPE, BINARY_ESCAPE);
				break;
			}

			default:
				throw new BaseException("Unhandled SQL column type for column '" + 
						column.getName() + "': " + sqlType);
			
		}
		return rw;
	}
	
}
