/**
	$Id: AbstractSignalHandler.java 6127 2012-09-14 07:28:35Z nicklas $

	Copyright (C) Authors contributing to this file.

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.signal;

import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

/**
	An abstract base class for signal handler implementations. This
	class will help an implementor of the {@link SignalHandler} interface
	to provide information about supported signals. The easiest way is to
	use the {@link #AbstractSignalHandler(Collection)} constructor.

	@author nicklas
	@version 2.6
	@base.modified $Date: 2012-09-14 09:28:35 +0200 (Fri, 14 Sep 2012) $
*/
public abstract class AbstractSignalHandler
	implements SignalHandler
{

	/**
		Holds the supported signals.
	*/
	private Set<Signal> supported;
	
	/**
		Create a new signal handler, that initially doesn't support any signals at 
		all.
	*/
	protected AbstractSignalHandler()
	{
		this(null);
	}

	/**
		Create a new signal handler that supports the given signals.
		
		@param supported A collection with the signals that are initially supported.
			More signals can be added with {@link #addSignal(Signal)}
	*/
	protected AbstractSignalHandler(Collection<Signal> supported)
	{
		this.supported = new HashSet<Signal>();
		if (supported != null) this.supported.addAll(supported);
	}
	
	/*
		From the SignalHandler interface
		-------------------------------------------
	*/
	@Override
	public Collection<Signal> getSupportedSignals()
	{
		return Collections.unmodifiableSet(supported);
	}
	@Override
	public boolean supports(Signal signal)
	{
		return supported != null && supported.contains(signal);
	}
	// -------------------------------------------

	/**
		Add a signal to the list of supported signals.
		@param signal The signal to add
	*/
	protected void addSignal(Signal signal)
	{
		if (signal != null) supported.add(signal);
	}
	/**
		Remove a signal from the list of supported signals.
		@param signal The signal to remove
	*/
	protected void removeSignal(Signal signal)
	{
		supported.remove(signal);
	}
	
}
