/*
	$Id$

	Copyright (C) 2020 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util;

import net.sf.basedb.core.Item;

/**
	Class for holding statisics about any-to-any links between items.
	
	@author nicklas
	@since 3.17
*/
public class AnyToAnyLinkStatistics 
	implements Comparable<AnyToAnyLinkStatistics>
{
	
	private int sampleSize;
	private String linkName;
	private Item targetType;
	private int count;
	
	public AnyToAnyLinkStatistics()
	{}
	
	/**
		Get the number of items that was sampled to generated the statistics.
	*/
	public int getSampleSize()
	{
		return sampleSize;
	}
	
	public void setSampleSize(int sampleSize)
	{
		this.sampleSize = sampleSize;
	}
	
	/**
		Get the name of the link.
	*/
	public String getLinkName()
	{
		return linkName;
	}
	
	public void setLinkName(String linkName)
	{
		this.linkName = linkName;
	}
	
	/**
		Get the item type of the target item for the link.
	*/
	public Item getTargetType()
	{
		return targetType;
	}
	
	public void setTargetType(Item targetType)
	{
		this.targetType = targetType;
	}
	
	/**
		Get the number of links that was found among the sampled items.
	*/
	public int getCount()
	{
		return count;
	}
	
	public void setCount(int count)
	{
		this.count = count;
	}

	/**
		Sort items by count, linkName and targetType.
	*/
	@Override
	public int compareTo(AnyToAnyLinkStatistics o) 
	{
		int result = count - o.count;
		if (result == 0) linkName.compareTo(o.linkName);
		if (result == 0) result = targetType.name().compareTo(o.targetType.name());
		return result;
	}

	@Override
	public int hashCode() 
	{
		return count + linkName.hashCode() + targetType.hashCode();
	}

	@Override
	public boolean equals(Object o) 
	{
		if (o == null) return false;
		if (!o.getClass().equals(getClass())) return false;
		AnyToAnyLinkStatistics other = (AnyToAnyLinkStatistics)o;
		return count == other.count && targetType == other.targetType && linkName.equals(other.linkName);
	}

	@Override
	public String toString() 
	{
		return linkName + "[" + targetType.name() + "]="+count;
	}
	
	
	/**
		Options to specify how the statistics should be generated.
	 * @author nicklas
	 *
	 */
	public static class Options
	{
		
		private int maxRandomSamples;
		private int minCount;
		private int minFraction;
		
		public Options()
		{
			this.maxRandomSamples = 500;
			this.minCount = 10;
			this.minFraction = 50;
		}
		
		/**
			Set the maximum number of items to use for gathering the statistics.
			The default value is 500. A higher value may affect performance.
		*/
		public void setMaxRandomSamples(int maxRandomSamples)
		{
			this.maxRandomSamples = maxRandomSamples;
		}
		
		public int getMaxRandomSamples()
		{
			return maxRandomSamples;
		}
		
		/**
			Do not return information about links with a count that is less
			than this value. The default value is 10. This option is ignored 
			if the fraction yields a lower count (sampleSize / fraction).
		*/
		public void setMinCount(int minCount)
		{
			this.minCount = minCount;
		}
		
		public int getMinCount()
		{
			return minCount;
		}
		
		/**
			Do not return information about links if they are found
			on less than the specified fraction of samples. The default
			value is 50, which means that links that are found in less than
			1 of 50 samples are ignored. This option is ignored if the
			min count is lower.
		*/
		public void setMinFraction(int minFraction)
		{
			this.minFraction = minFraction;
		}
		
		public int getMinFraction()
		{
			return minFraction;
		}
		
	}
}
