package net.sf.basedb.util.charset;

import java.io.IOException;

/**
	A simple string detector implementation that works with two strings.
	It is designed to be used to detect the encoding in tabular data files
	using one of the ISO-8859-N encodings or similar that are not not possible
	to separate techically.
	
	The data file is expected to contain a header line were at least one header 
	column has a name with non-ASCII characters.
	
	For each line of data if will first check if the 'ifFound' string 
	can be found. If not, it will return null to request more data.
	
	If the 'ifFound' string is found, it will continue to see if the 'thenMatch'
	string is also present. If so, TRUE is returned to indicate a successful
	encoding match, otherwise FALSE is return to indicate an incorrect encoding.
	
	Note that the two strings need to be selected wisely. The 'ifFound' string should
	typcially be an ASCII-only string and 'thenMatch' a string with one or more non-ASCII
	characters.

	For example, if the file header is: <code>Namn{tab}Ålder</code>,
	we could use 'ifFound=Namn' and 'thenMatch=Ålder'.
	
	If the entire file is parsed without finding the 'ifFound' string, the {@link #eof(int)} method
	will return false.
	
	@author nicklas
	@since 3.15
*/
public class SimpleStringDetector 
	implements StringDetector 
{

	private final String ifFound;
	private final String thenMatch;
	
	public SimpleStringDetector(String ifFound, String thenMatch) 
	{
		this.ifFound = ifFound;
		this.thenMatch = thenMatch;
	}
	
	@Override
	public boolean checkLine(int lineNo, String line)
		throws IOException
	{
		if (line == null || !line.contains(ifFound)) return false;
		if (line.contains(thenMatch)) return true;
		throw new IOException("Found '"+ifFound+"' on line " + lineNo + " but not '" + thenMatch + "': "+line);
	}

	@Override
	public void eof(int parsedLines) 
		throws IOException
	{
		throw new IOException("Could not find '"+ifFound + "' after " + parsedLines + " lines");
	}

}
