/**
	$Id$

	Copyright (C) 2023 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.extensions.logging;

/**
	A simple logger that services can use to send log message to the BASE
	core. An instance is handed out to the service by the {@link 
	ServiceControllerAction#start(ExtensionsLogger)} method, which the service
	can use as long as it is running.
	<p>
	Note that the BASE core will automatically send INFO-level messages
	when the service is started and stopped.
	<p>
	The log is kept in memory only and will not survive server restarts.
	The number of log entries is also limited. When the limit is reached
	the oldest entries are automatically removed from the log.
	
	@author nicklas
	@since 3.19.8
*/
public class ExtensionsLogger 
{
	private final ExtensionsLog log;
	private final org.slf4j.Logger otherLogger;
	
	ExtensionsLogger(ExtensionsLog log)
	{
		this(log, null);
	}
	
	ExtensionsLogger(ExtensionsLog log, org.slf4j.Logger otherLogger)
	{
		this.log = log;
		this.otherLogger = otherLogger;
	}
	
	/**
		Wrap a generic SLF4J logger with this logger. Messages
		that are submitted to the logger created by this method are
		also forwarded to the SLF4J logger.
	*/
	public ExtensionsLogger wrap(org.slf4j.Logger otherLogger)
	{
		return new ExtensionsLogger(this.log, otherLogger);
	}
	
	/**
		Check if DEBUG logging has been enabled or not. Debug logging
		can be enabled either by the {@link ExtensionsLog#setDebugEnabled(boolean)}
		method or by a wrapped {@link Logger}.
	*/
	public boolean isDebugEnabled()
	{
		return log.isDebugEnabled() || otherLogger != null && otherLogger.isDebugEnabled();
	}
	
	/**
		Check if TRACE logging has been enabled or not. Trace logging is enabled
		only if both DEBUG logging and {@link ExtensionsLog#isCopyingToStdout()}
		are enabled or if a wrapped {@link Logger} has it enabled.
	*/
	public boolean isTraceEnabled()
	{
		return (log.isCopyingToStdout() && isDebugEnabled()) || 
			(otherLogger != null && otherLogger.isTraceEnabled());
	}
	
	/**
		Log a message at the specified level.
	*/
	public void log(LogLevel level, String msg)
	{
		log.add(new LogEntry(level, msg, null));
		if (otherLogger != null) level.logTo(otherLogger, msg, null);
	}
	
	/**
		Log a message at the specified level with an optional stacktrace.
	*/
	public void log(LogLevel level, String msg, Throwable stacktrace)
	{
		log.add(new LogEntry(level, msg, stacktrace));
		if (otherLogger != null) level.logTo(otherLogger, msg, stacktrace);
	}
	
	/**
		Log an INFO message.
	*/
	public void info(String msg)
	{
		log(LogLevel.INFO, msg);
	}
	/**
		Log an INFO message with an optional stacktrace.
	*/
	public void info(String msg, Throwable stacktrace)
	{
		log(LogLevel.INFO, msg, stacktrace);
	}
	
	/**
		Log a WARNING message.
	*/
	public void warn(String msg)
	{
		log(LogLevel.WARNING, msg);
	}
	/**
		Log a WARNING message with an optional stacktrace.
	*/
	public void warn(String msg, Throwable t)
	{
		log(LogLevel.WARNING, msg, t);
	}
	
	/**
		Log an ERROR message.
	*/
	public void error(String msg)
	{
		log(LogLevel.ERROR, msg);
	}
	/**
		Log an ERROR message with an optional stacktrace.
	*/
	public void error(String msg, Throwable t)
	{
		log(LogLevel.ERROR, msg, t);
	}
	
	/**
		Log a DEBUG message.
	*/
	public void debug(String msg)
	{
		log(LogLevel.DEBUG, msg);
	}
	/**
		Log a DEBUG message with an optional stacktrace.
	*/
	public void debug(String msg, Throwable t)
	{
		log(LogLevel.DEBUG, msg, t);
	}
	
	/**
		Log a TRACE message with the related SLF4J logger. 
		If no SLF4J logger is related 
		(see {@link #wrap(org.slf4j.Logger)}) all TRACE messages
		will be ignored.
	*/
	public void trace(String msg)
	{
		log(LogLevel.TRACE, msg);
	}
	/**
		Log a TRACE message with an optional stacktrace.
		If no SLF4J logger is related 
		(see {@link #wrap(org.slf4j.Logger)}) all TRACE messages
		will be ignored.
	*/
	public void trace(String msg, Throwable t)
	{
		log(LogLevel.TRACE, msg, t);
	}
	
}