/**
	$Id: PluginInfo.java 5821 2011-10-24 10:47:31Z nicklas $

	Copyright (C) 2011 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.extensions.xml;

import java.util.Comparator;
import java.util.HashMap;
import java.util.Map;

import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.PluginDefinition;
import net.sf.basedb.core.plugin.About;
import net.sf.basedb.core.plugin.Plugin.MainType;

/**
	Object for holding information about a plug-in definition as
	it is loaded from the extensions definition file. This class
	has no effect on the actual installed plug-ins. To get
	more information about what is installed and what is not,
	call {@link PluginDefinition#checkInstallation(DbControl, java.util.Collection)}.

	@author Nicklas
	@since 3.0
	@base.modified $Date: 2011-10-24 12:47:31 +0200 (Mon, 24 Oct 2011) $
*/
public class PluginInfo
{

	
	/**
		Comparator that sort plug-ins according to their names.
		If two names are equal, the sort is also done on the class 
		name.
	*/
	public static final Comparator<PluginInfo> NAME_COMPARATOR = 
		new Comparator<PluginInfo>()
		{
			@Override
			public int compare(PluginInfo o1, PluginInfo o2)
			{
				String name1 = o1.getAbout() == null ? null : o1.getAbout().getName();
				String name2 = o2.getAbout() == null ? null : o2.getAbout().getName();
				int result = name1 != null && name2 != null ? name1.compareTo(name2) : 0;
				if (result == 0) result = o1.getClassName().compareTo(o2.getClassName());
				return result;
			}
		};

	
	private final String id;
	private About about;
	private MainType type;
	private String className;
	private Map<String, String> properties;
	
	private int internalId;
	private boolean disabled;
	private boolean notFound;
	
	/**
		Create a new information object.
	*/
	public PluginInfo(String id)
	{
		this.id = id;
	}
	
	/**
		Get the class name of the plug-in.
	*/
	public String getClassName()
	{
		return className;
	}
	public void setClassName(String className)
	{
		this.className = className;
	}
	
	/**
		Get information about the authors of the plug-in.
	*/
	public About getAbout()
	{
		return about;
	}
	public void setAbout(About about)
	{
		this.about = about;
	}
	
	/**
		Get the main type of the plug-in. This information may
		not be available for plug-ins that has not been installed.
	*/
	public MainType getType()
	{
		return type;
	}
	
	public void setType(MainType type)
	{
		this.type = type;
	}
	
	public void setProperty(String name, String value)
	{
		if (properties == null) properties = new HashMap<String, String>();
		properties.put(name, value);
	}
	
	public String getProperty(String name)
	{
		return properties == null ? null : properties.get(name);
	}
	
	/**
		Get the internal if of this plug-in. This information is
		only available after {@link PluginDefinition#checkInstallation(DbControl, java.util.Collection)}
		has been called.
	*/
	public int getInternalId()
	{
		return internalId;
	}
	
	public void setInternalId(int internalId)
	{
		this.internalId = internalId;
	}
	
	/**
		Is this plug-in installed or not?
	*/
	public boolean isInstalled()
	{
		return internalId != 0;
	}
	
	/**
		Is this plug-in enabled or not?
	*/
	public boolean isDisabled()
	{
		return disabled;
	}
	
	public void setDisabled(boolean disabled)
	{
		this.disabled = disabled;
	}
	
	/**
		Does this plug-in exists on the server or not?
	*/
	public boolean isNotFound()
	{
		return notFound;
	}
	
	public void setNotFound(boolean notFound)
	{
		this.notFound = notFound;
	}
}
