/**
	$Id: ResetMetadataRenderer.java 8117 2023-02-22 07:58:38Z nicklas $

	Copyright (C) 2011 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.fileset;

import java.util.ArrayList;
import java.util.List;

import net.sf.basedb.core.signal.ThreadSignalHandler;
import net.sf.basedb.util.extensions.AbstractRenderer;
import net.sf.basedb.util.extensions.Extension;

/**
	Renderer implementation for resetting metadata for the members of a file set.

	@author Nicklas
	@since 3.0
	@base.modified $Date: 2023-02-22 08:58:38 +0100 (Wed, 22 Feb 2023) $
*/
public class ResetMetadataRenderer
	extends AbstractRenderer<ValidationAction>
{

	private final List<ValidatingFileSetMember> members;

	/**
		Create a new renderer instance.
	
		@param members A list with the members of the file set that should have
			their metadata reset
	 */
	public ResetMetadataRenderer(List<ValidatingFileSetMember> members)
	{
		this.members = members;
	}
	
	/*
		From the Renderer interface
		---------------------------
	*/
	@Override
	public void render(ValidationAction action, Extension<? extends ValidationAction> ext)
	{
		ThreadSignalHandler.checkInterrupted();
		
		// Check if the action accepts any file(s)
		List<ValidatingFileSetMember> accepted = new ArrayList<ValidatingFileSetMember>();
		for (ValidatingFileSetMember member : members)
		{
			Accept accept = member.isAcceptedByValidator(action);
			if (accept == Accept.VALIDATE_IMMEDIATELY)
			{
				// Reset metadata immediately
				action.resetMetadata();
			}
			else if (accept == Accept.VALIDATE_LATER)
			{
				// File will be used in batch later
				accepted.add(member);
			}
		}
		
		// Are there any files waiting for validation?
		if (accepted.size() == 0) return;

		// Reset metadata for the remaining files
		action.resetMetadata();
	}
	// ---------------------------------
}
