/**
	$Id: GtfInputStream.java 6898 2015-05-12 09:27:08Z nicklas $

	Copyright (C) 2011 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.gtf;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import net.sf.basedb.util.parser.FlatFileParser;

/**
	Input stream implementation that reads from a GTF file and converts it to
	a simple tab-separated file with a single line of column headers. This is
	useful since it means that we can use the regular {@link FlatFileParser} 
	and other tools for parsing the resulting stream. The first line in the
	file is used a template line. The first 8 columns are fixed. The 9th column
	contains attributes as key/value pairs, which are converted to additional
	columns in the output. The GTF specification require that <code>gene_id</code>
	and <code>transcript_id</code> are present, which means that the output will
	contain at least 10 columns. Subsequent lines are parsed in the same way and
	attributes are lined up with the first line. Note that any attributes
	that are not present in the first line are skipped. The parser also has an
	option to skip lines with a <code>transcript_id+seqname</code> that is not unique.
	Normally, a GTF file will contain multiple entries with the same id:s, but
	in most cases we are not interested in this when importing data to BASE. 
	This option also remove the feature, start, end, score, strand and frame
	columns from the output. Lines that can't be split into at least 9 columns 
	(eg. comment lines starting with #) are ignored and forwarded without modification.

	@author Nicklas
	@since 3.0
	@base.modified $Date: 2015-05-12 11:27:08 +0200 (Tue, 12 May 2015) $
*/
public class GtfInputStream
	extends InputStream
{
	
	private final InputStream master;
	private final BufferedReader reader;
	private final Charset charset;
	private final Pattern ATTRIBUTE_PATTERN;
	
	// Buffer each line that we read from the GTF file
	private byte[] buffer;
	// The index of the next byte to return from the buffer
	private int index;

	// Number of lines that we have read from the GTF file
	private int lineNum;
	
	// List of attributes in the GTF file in the order they appear
	// on the first line. gene_id and transcript_id are required
	// so we remember the index they have
	private Attribute[] attributes;
	private int geneIdIndex;
	private int transcriptIdIndex;
	
	// For storing id:s that we have seen and may be skipped
	private final boolean skipRepeatedTranscriptIds;
	private final Set<String> transcriptIds;
	
	/**
		Create a new input stream reading from the master.
		@param master The master input stream
		@param charset The character set used in the file
		@param skipRepeatedTranscriptIds TRUE to skip lines with non-unique
			values for transcript_id+seqname
	 */
	public GtfInputStream(InputStream master, String charset, boolean skipRepeatedTranscriptIds)
		throws IOException
	{
		this.master = master;
		this.charset = Charset.forName(charset == null ? "ISO-8859-1" : charset);
		this.skipRepeatedTranscriptIds = skipRepeatedTranscriptIds;
		this.transcriptIds = new HashSet<String>();
		this.reader = new BufferedReader(new InputStreamReader(master, this.charset));
		this.ATTRIBUTE_PATTERN = Pattern.compile("([^ ]+) (([^\";]*)|(\"[^\"]*\"));");
		this.buffer = readMore();
	}

	/*
		From the InputStream class
		--------------------------
	*/
	@Override
	public int read() 
		throws IOException
	{
		if (buffer == null) return -1;
		if (index >= buffer.length)
		{
			// read next data line
			readMore();
			index = 0;
			if (buffer == null) return -1;
		}
		
		// Important! Mask with 0xff to get values in range (0,255), NOT (-128,127)
		return buffer[index] & 0xff;
	}

	@Override
	public int read(byte[] b)
		throws IOException
	{
		return read(b, 0, b.length);
	}

	@Override
	public int read(byte[] b, int off, int len) 
		throws IOException
	{
		if (buffer == null) return -1;
		if (index >= buffer.length)
		{
			// read next data line
			buffer = readMore();
			index = 0;
			if (buffer == null) return -1;
		}
		if (len > buffer.length - index)
		{
			len = buffer.length - index;
		}
		System.arraycopy(buffer, index, b, off, len);
		index += len;
		return len;
	}

	@Override
	public int available() 
		throws IOException
	{
		return buffer == null ? 0 : buffer.length - index;
	}

	@Override
	public void close()
		throws IOException
	{
		master.close();
	}

	@Override
	public boolean markSupported()
	{
		return false;
	}

	@Override
	public synchronized void reset() 
		throws IOException
	{
		throw new IOException("reset() is not supported");
	}
	// -----------------------------------------

	/**
		Get the number of lines parsed so far.
	*/
	public int getNumLines()
	{
		return lineNum;
	}
	
	/**
		Get the number of unique transcript ids found so far.
	*/
	public int getNumUniqueTranscriptIds()
	{
		return transcriptIds.size();
	}
	
	/**
		Read more data from the GTF file. Typically one additional
		line is read and stored in the buffer. Do not call this method
		unless it is certain that the existing buffer has been completely
		read by the reader of this input stream.
	*/
	private byte[] readMore()
		throws IOException
	{
		String[] line;
		StringBuffer sb = new StringBuffer();
		do
		{
			line = getNextLine();
			if (line == null) 
			{
				buffer = null;
				return null;
			}
			
			// Parse attributes in 9th column
			if (line.length >= 9)
			{
				// Generate headers if this is the first line with data
				boolean generateHeaders = attributes == null;
				parseAttributes(line[8]);
				if (generateHeaders)
				{
					if (skipRepeatedTranscriptIds)
					{
						sb.append("<seqname>\t<source>");
					}
					else
					{
						sb.append("<seqname>\t<source>\t<feature>\t<start>\t<end>\t<score>\t<strand>\t<frame>");
					}
					for (Attribute attribute : attributes)
					{
						sb.append("\t<").append(attribute.key).append(">");
					}
					sb.append("\n");
				}
				
				// Break out of the loop as soon as we see a new transcript id or if we are including repeated ids
				String id = attributes[transcriptIdIndex].value + '@' + line[0];
				if (transcriptIds.add(id) || !skipRepeatedTranscriptIds)
				{
					appendLine(sb, line, attributes);
					break;
				}
			}
			else
			{
				// Append line as it is
				appendLine(sb, line, null);
				break;
			}
		} while (skipRepeatedTranscriptIds);
		
		// Convert to byte[]
		return sb.toString().getBytes(charset);
	}
	
	/**
		Read the next line from the GTF file and split on tab character.
	*/
	private String[] getNextLine()
		throws IOException
	{
		String line = reader.readLine();
		if (line == null) return null;
		lineNum++;
		String[] columns = line.split("\\t", 10);
		return columns;
	}
	
	/**
		Parse attributes from the given template string. The first time
		this method is called all attributes are accepted and their order
		is remembered. Subsequent calls accept only values for the remembered
		attributes.
	*/
	private void parseAttributes(String template)
		throws IOException
	{
		Matcher m = ATTRIBUTE_PATTERN.matcher(template);
		
		if (attributes == null)
		{
			// First time
			List<Attribute> tmp = new ArrayList<Attribute>();
			while (m.find())
			{
				String key = m.group(1);
				String value = m.group(2);
				tmp.add(new Attribute(key, value));
			}
			geneIdIndex = tmp.indexOf(Attribute.GENE_ID);
			if (geneIdIndex == -1)
			{
				throw new IOException("Required attribute 'gene_id' not found, at line: " + lineNum);
			}
			transcriptIdIndex = tmp.indexOf(Attribute.TRANSCRIPT_ID);
			if (transcriptIdIndex == -1)
			{
				throw new IOException("Required attribute 'transcript_id' not found, at line: " + lineNum);
			}
			attributes = tmp.toArray(new Attribute[tmp.size()]);
		}
		else
		{
			// Subsequent times
			int startAt = 0;
			while (m.find())
			{
				String key = m.group(1);
				String value = m.group(2);
				for (int i = startAt; i < attributes.length; ++i)
				{
					if (attributes[i].key.equals(key))
					{
						attributes[i].value = value;
						// If the order of the attributes is the same on all lines
						// we don't have to look from [0] to [length-1] for all attributes
						if (i == startAt) startAt++;
						break;
					}
				}
			}
			if (attributes[geneIdIndex].value == null)
			{
				throw new IOException("Required attribute 'gene_id' not found, at line: " + lineNum);
			}
			if (attributes[transcriptIdIndex].value == null)
			{
				throw new IOException("Required attribute 'transcript_id' not found, at line: " + lineNum);
			}
		}
	}
	
	/**
		Append columns to the buffer and separate each with a tab.
		If attributes are given, the first 8 (or 2 if skipRepeatedTranscriptIds=true)
		columns are appended, then each of the attributes are appended.
		If no attributes are given, all columns are copied as they are.
		
		@param sb The buffer to append to
		@param columns The regular columns (must be at least 8)
		@param attr The attributes to add
	*/
	private StringBuffer appendLine(StringBuffer sb, String[] columns, Attribute[] attr)
	{
		// First 8 columns are copied with tab separator
		sb.append(columns[0]);
		int end = attr == null ? columns.length : (skipRepeatedTranscriptIds ? 2 : 8);
		for (int i = 1; i < end; ++i)
		{
			sb.append("\t").append(columns[i]);
		}
		if (attr != null)
		{
			// Then follows the attributes
			for (Attribute attribute : attr)
			{
				sb.append("\t").append(attribute.value);
				attribute.value = null;
			}
			if (columns.length == 10)
			{
				sb.append("\t").append(columns[9]);
			}
		}
		sb.append("\n");
		return sb;
	}
	
	
	static class Attribute
	{
		static final Attribute GENE_ID = new Attribute("gene_id", null);
		static final Attribute TRANSCRIPT_ID = new Attribute("transcript_id", null);
		
		final String key;
		String value;
	
		Attribute(String key, String value)
		{
			this.key = key;
			this.value = value;
		}
		
		@Override
		public int hashCode()
		{
			return key.hashCode();
		}
		@Override
		public boolean equals(Object obj)
		{
			if (this == obj) return true;
			if (obj == null) return false;
			if (obj.getClass() != Attribute.class) return false;
			Attribute o = (Attribute)obj;
			return this.key.equals(o.key);
		}
		@Override
		public String toString()
		{
			return key + "=" + value;
		}
		
		
	}
	
}
