/*
	$Id: ReporterFunction.java 6875 2015-04-20 09:08:18Z nicklas $

	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.jep;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.DynamicQuery;
import net.sf.basedb.core.RawDataType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Metadata;
import net.sf.basedb.core.data.ReporterData;
import net.sf.basedb.core.query.Dynamic;
import net.sf.basedb.core.query.Expression;
import net.sf.basedb.util.BioAssaySetUtil;

import java.util.Stack;

import org.nfunk.jep.Node;
import org.nfunk.jep.ParseException;


/**
	A JEP function class that adds a <code>rep(string)</code> or <code>mrep(string)</code>
	function to a  JEP expression parser. The function will look up the value of the reporter 
	property with the given name. For example: 
	<code>rep('sequence')</code>
	<p>
	If the <code>mrep</code> variant is used, the master reporter table is used,
	otherwise the actual query can decide if it should use cloned reporter information
	or not.
	<p>
	To be able to use this function it must be registered with the JEP
	parser and, before the expression is evaluated, a reporter object must be set.
	For example we can evaluate an expression for every reporter object:
	
	<pre class="code">
DbControl dc = ...
RawBioAssay assay = ...
String expression = "rep('sequence')";
ReporterFunction rep = new RepFunction();
JEP jep = JepUtil.newJep(expression, rep);
DataResultIterator&lt;RawData&gt; result = 
   assay.getRawData().iterate(dc);
while (result.hasNext())
{
   rep.setReporter(result.next().getReporter());
   String value = jep.getValue();
   // Do something with the value
}
result.close();
</pre>

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2015-04-20 11:08:18 +0200 (Mon, 20 Apr 2015) $
	@see Jep
	@see BioAssaySetUtil#createJepExpression(DbControl, String, DynamicQuery)
*/
public class ReporterFunction
	implements JepExpressionFunction
{

	private final Metadata<ReporterData> metaData;
	private final DbControl dc;
	private final boolean master;
	private int numParameters;
	private ReporterData reporter;

	/**
		Create a new instance of this function. The new instance cannot be used
		to dynamically evaluate expressions. It should only be used for converting
		JEP formulas to {@link Expression}:s.
		@see Jep#formulaToExpression(String, JepFunction[])
	*/
	public ReporterFunction()
	{
		this(false);
	}
	
	/**
		Create a new instance of this function. The new instance cannot be used
		to dynamically evaluate expressions. It should only be used for converting
		JEP formulas to {@link Expression}:s. If 
		
	 	@since 3.1
		@param master TRUE to force using the master reporter table, FALSE to let
			the query decide
	*/
	public ReporterFunction(boolean master)
	{
		this.dc = null;
		this.metaData = null;
		this.master = master;
	}
	
	/**
		Create a new instance of this function working with reporters.
	*/
	public ReporterFunction(DbControl dc, RawDataType rawDataType)
	{
		this.dc = dc;
		this.metaData = Metadata.getInstance(ReporterData.class);
		this.master = false;
	}
	
	/*
		From the JepFunction interface
		-------------------------------------------
	*/
	/**
		@return The string "rep" or "mrep"
	*/
	@Override
	public String getFunctionName()
	{
		return master ? "mrep" : "rep";
	}
	// -------------------------------------------
	/*
		From the JepExpressionFunction interface
		-------------------------------------------
	*/
	/**
		Use the {@link Dynamic#reporter(String)} method to create an
		expression referencing a reporter property.
	*/
	@Override
	public Expression toExpression(Node node)
	{
		int numChildren = node.jjtGetNumChildren();
		if (numChildren != 1)
		{
			throw new BaseException("Invalid number of expressions for 'rep' function: " + numChildren);
		}
		return Dynamic.reporter(Jep.nodeToString(node.jjtGetChild(0)), master ? false : null);
	}
	// -------------------------------------------
	/*
		From the PostfixMathCommandI interface
		-------------------------------------------
	*/
	/**
		@return Always 1
	*/
	@Override
	public int getNumberOfParameters()
	{
		return 1;
	}
	@Override
	public void setCurNumberOfParameters(int n)
	{
		this.numParameters = n;
	}
	@Override
	public boolean checkNumberOfParameters(int n)
	{
		return n == 1;
	}
	@Override
	@SuppressWarnings({"unchecked", "rawtypes"})
	public void run(Stack stack)
		throws ParseException
	{
		if (stack == null || stack.empty()) 
		{
			throw new ParseException("Stack is empty");
		}
		Object propertyName = stack.pop();
		if (propertyName instanceof String)
		{
			stack.push(raw((String)propertyName));
		}
		else
		{
			throw new ParseException("Invalid parameter type: " + propertyName + "; expected string");
		}
	}
	// -------------------------------------------

	/**
		Set a new reporter object that will be used the next time the
		JEP expression is evaluated.
		@param reporter The reporter object to use
	*/
	public void setReporter(ReporterData reporter)
	{
		this.reporter = reporter;
	}
	/**
		Get the current reporter object used when evaluating the JEP expression.
	*/
	public ReporterData getReporter()
	{
		return reporter;
	}
	
	/**
		Get the value of the specified property of the current raw data object.
		This method uses Hibernate metadata to find the value.
		@param propertyName The name of the raw data property
	 	@return an java.lang.Object 
	 	@throws ParseException If the argument is null or 'reporter' not specified. 
	*/
	public Object raw(String propertyName)
		throws ParseException
	{
		if (reporter == null || propertyName == null)
		{
			throw new ParseException("No reporter object has been specified for function rep('"+propertyName+"')");
		}
		return metaData.getPropertyPath(propertyName, false).getValue(dc, reporter);
	}
}
