/**
	$Id$

	Copyright (C) 2015 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.listable;

import java.util.Set;

import net.sf.basedb.core.BioSource;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;

/**
	Source item transformer that transform sample items to their parent
	biosources. This implementation will only transform to the sample:s
	immediate parent biosources. Grandparents are not included. To do this, 
	combine this transformer with a {@link SampleToParentSampleTransformer}.
	
	@author Nicklas
	@since 3.5
*/
public class SampleToBioSourceTransformer 
	extends AbstractSourceItemTransformer 
{

	private boolean pushOnly;
	
	public SampleToBioSourceTransformer()
	{
		super(Item.SAMPLE, Item.BIOSOURCE);
	}
	
	/**
		Create a new sample to biosource transformer that only load parent biosources
		to sample that has a subtype with "push annotations" set. This functionality
		should only be used in the first transformation step.
		@see PushToParentsTransformer
		@since 3.16
	*/
	public SampleToBioSourceTransformer(boolean childrensThatPushOnly)
	{
		this();
		this.pushOnly = childrensThatPushOnly;
	}
	
	/**
		Load the ID of all biosources that are immediate parents to the
		given set of samples.
	*/
	@Override
	public Set<Integer> transform(TransformContext context, Set<Integer> source) 
	{
		context.collect(getSourceItemType(), source);
		ItemQuery<BioSource> query = BioSource.getQuery();
		query.setIncludes(context.getInclude());
		query.join(Hql.innerJoin("children", "ch"));
		query.restrict(
			Restrictions.in(
				Hql.alias("ch"), 
				Expressions.parameter("samples")
			)
		);
		if (pushOnly)
		{
			query.join(Hql.innerJoin("ch", "itemSubtype", "st"));
			query.restrict(Restrictions.eq(Hql.property("st", "pushAnnotations"), Expressions.bool(true)));
		}
		return context.avoid(getTargetItemType(), safeIdList(context.getDbControl(), query, "samples", source));
	}

}
