/**
	$Id$

	Copyright (C) 2012 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.loader;

import net.sf.basedb.core.BioPlate;
import net.sf.basedb.core.BioWell;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Extract;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.ItemResultIterator;
import net.sf.basedb.core.MeasuredBioMaterial;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.core.Sample;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Orders;
import net.sf.basedb.util.overview.Node;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.node.BioWellNameGenerator;
import net.sf.basedb.util.overview.node.ChildNodeDirection;
import net.sf.basedb.util.overview.node.NodeFactory;
import net.sf.basedb.util.overview.validator.NodeValidator;

/**
	Node loader implementation that load biomaterial on a bioplate.
	This loader is only user in the forward-loading direction from a
	bioplate to the biomaterial. Note that the loader returns biomaterial nodes,
	NOT biowell nodes. It can be used as a root node or in the reverse direction.
	
	@author nicklas
	@since 3.2
*/
public class BioWellLoader 
	extends BasicItemNodeLoader<MeasuredBioMaterial> 
{
	
	public BioWellLoader()
	{
		super(Item.BIOWELL, DENY_ROOT_NODE, null);
	}

	
	/*
		From the NodeLoader interface
		------------------------------
	*/
	/**
		Create forward-loading biomaterial nodes from a given bioplate
		The biomaterial nodes are created directly as child nodes to the
		bioplate. No intermediate folder node is used so this method return 
		null.
		@return Always null (but child nodes are added to the parent node)
	*/
	@Override
	public Node createForwardNode(DbControl dc, OverviewContext context, Node parentNode)
	{
		Node returnNode = null;
		Item parentType = parentNode.getItemType();
		if (parentType == Item.BIOPLATE)
		{
			returnNode = createForwardNode((BioPlate)parentNode.getItem(dc), dc, context, parentNode);
		}
		return returnNode;
	}
	
	/**
		Delegate loading of child nodes to the actual item loader that this node
		is representing.
	*/
	@Override
	public boolean loadChildNodes(DbControl dc, OverviewContext context, Node node)
	{
		return getNodeLoader(context, node.getItemType()).loadChildNodes(dc, context, node);
	}
	// --------------------------------

	/**
		Create all-loading biomaterial nodes for each non-empty well on the bioplate.
	*/
	@SuppressWarnings("unchecked")
	private Node createForwardNode(BioPlate plate, DbControl dc, OverviewContext context, Node plateNode)
	{
		// Name generate that prefixes each name with the well coordinate
		BioWellNameGenerator nameGenerator = new BioWellNameGenerator();
		Node folderNode = null;
		
		// Validator and node factory for wells with samples
		NodeValidator<Sample> sampleNodeValidator = (NodeValidator<Sample>)getNodeValidatorFactory(context).createNodeValidator(Item.SAMPLE);
		NodeFactory<Sample> sampleNodeFactory = new NodeFactory<Sample>(dc, context, sampleNodeValidator, nameGenerator);

		// Validator and node factory for wells with extracts
		NodeValidator<Extract> extractNodeValidator = (NodeValidator<Extract>)getNodeValidatorFactory(context).createNodeValidator(Item.EXTRACT);
		NodeFactory<Extract> extractNodeFactory = new NodeFactory<Extract>(dc, context, extractNodeValidator, nameGenerator);

		// Find all used wells
		ItemQuery<BioWell> query = context.initQuery(plate.getBioWells(), Orders.asc(Hql.property("row")), Orders.asc(Hql.property("column")));
		query.join(Hql.innerJoin(null, "bioMaterial", "bm", true)); // Important to use inner join for filtering
		ItemResultIterator<BioWell> it = query.iterate(dc);
		while (it.hasNext())
		{
			BioWell well = it.next();
			nameGenerator.setCurrentWell(well); // The name generator must know about the current well
			MeasuredBioMaterial bioMaterial = null;
			boolean denied = false;
			try
			{
				bioMaterial = well.getBioMaterial();
			}
			catch (PermissionDeniedException ex)
			{
				denied = true;
			}
			
			if (folderNode == null)
			{
				folderNode = new Node("biomaterial", "Biomaterial", plateNode, ChildNodeDirection.FORWARD);
			}
			if (well.getBioMaterialType() == Item.SAMPLE)
			{
				// Create sample node
				Sample sample = (Sample)bioMaterial;
				createItemNode(sampleNodeFactory, sample, well, denied, folderNode, ChildNodeDirection.ALL);
			}
			else
			{
				// Create extract node
				Extract extract = (Extract)bioMaterial;
				createItemNode(extractNodeFactory, extract, well, denied, folderNode, ChildNodeDirection.ALL);
			}
		}
		return folderNode;
	}

}
