/**
	$Id: NodeLoader.java 5818 2011-10-21 11:15:41Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.loader;

import net.sf.basedb.core.DbControl;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.node.ChildNodeDirection;
import net.sf.basedb.util.overview.Node;

/**
	A node loader is an object that knows how to create a node for some
	specific item and how to load forward and reverse nodes for related
	items.
	<p>
	
	NOTE! This interface is not fixed on the implementor side. We reserve
	the right to add new methods and functionality to this interface without
	warning. We recommend that all implementations are subclasses of {@link
	AbstractNodeLoader} since this class will get default implementation of
	all new funcationality.
	
	@author Nicklas
	@version 2.10
	@base.modified $Date: 2011-10-21 13:15:41 +0200 (Fri, 21 Oct 2011) $
*/
public interface NodeLoader<I>
{
	
	/**
		Create a root node for the given item. A root node is a node with
		no parent. Typically, the direction of the created node is
		{@link ChildNodeDirection#ALL} so that both forward and reverse
		child nodes are loaded.
		
		@param dc The DbControl to use for database access
		@param context The overview context
		@param item The root item
		@return A root node
		@throws UnsupportedOperationException If the current item can't
			be used as a root node
	*/
	public Node createRootNode(DbControl dc, OverviewContext context, I item);

	/**
		Create a forward-loading (from parent to child) node for the given parent node. 
		The loader should use the information on the parent node to get the correct item 
		(of type I) and create a node for that item. If the parent can have many child nodes,
		the loader should first create a folder-type node 
		and put the children inside that.
		<p>
		The direction of the node(s) should usually be {@link ChildNodeDirection#FORWARD}. 
		In case there is an error (permission denied, etc.) {@link 
		ChildNodeDirection#NONE} should be used. If there is no child item null
		should be returned.
		<p>
		Errors, missing items, etc. should be reported as failures to the {@link
		OverviewContext}.
		
		@param dc The DbControl to use for database access
		@param context The overview context
		@param parentNode The parent node
		@return A forward-loading node (may be a folder-type node with many subnodes),
			or null
		@throws UnsupportedOperationException If the current item can't
			be used as a forward-loading node
	*/
	public Node createForwardNode(DbControl dc, OverviewContext context, Node parentNode);
	
	/**
		Create a reverse-loading (from child to parent) node for the given child node. 
		The loader should use the information on the child node to get the correct item 
		(of type I) and create a node for that item. If the child can have many parent nodes,
		the loader should first create a folder-type node 
		and put the parents inside that.
		<p>
		The direction of the node(s) should usually be {@link ChildNodeDirection#REVERSE}. 
		In case there is an error (permission denied, etc.) {@link 
		ChildNodeDirection#NONE} should be used. If there is no parent item null
		should be returned.
		<p>
		Errors, missing items, etc. should be reported as failures to the {@link
		OverviewContext}.
		
		@param dc The DbControl to use for database access
		@param context The overview context
		@param childNode The child node
		@return A reverse-loading node (may be a folder-type node with many subnodes),
			or null
		@throws UnsupportedOperationException If the current item can't
			be used as a reverse-loading node
	*/
	public Node createReverseNode(DbControl dc, OverviewContext context, Node childNode);
	
	/**
		Create a property node for the given parent node. A property node is a node
		that is not on the main parent-child path. It is typically a node that
		can't be used a root node (for example, a protocol or software). If the
		parent node can have many properties of this type the loader can decide if
		it should first create a folder-type node and put the children inside the
		folder, or if all child nodes should be created directly in the 
		parent node. In the first case, the folder node should be returned. In
		the latter case, null should be returned.
		<p>
		The direction of the node(s) should usually be {@link ChildNodeDirection#NONE}, but
		it may also be {@link ChildNodeDirection#PROPERTY} in case the property has sub-properties.
		One example is the {@link ProtocolLoader} which loads the
		protocol parameters as child nodes.
		<p>
		In case there is an error (permission denied, etc.) {@link 
		ChildNodeDirection#NONE} should be used. If there is no property item null
		should be returned (but this may also indicate that more than one node
		was created).
		<p>
		Errors, missing items, etc. should be reported as failures to the {@link
		OverviewContext}.
		
		@param dc The DbControl to use for database access
		@param context The overview context
		@param parentNode The parent node
		@return A node (may be a folder-type node with many subnodes),
			or null
		@throws UnsupportedOperationException If the current item can't
			be used as a property node
	*/
	public Node createPropertyNode(DbControl dc, OverviewContext context, Node parentNode);
	
	/**
		Load child nodes of the current node unless they have been loaded already.
		Note that which child nodes to load depends on the {@link Node#getChildNodeDirection()}
		of the node.
		
		@param dc The DbControl to use for database access
		@param context The overview context
		@param node The node
		@return TRUE if the call resulted in new nodes being loaded, FALSE otherwise
	*/
	public boolean loadChildNodes(DbControl dc, OverviewContext context, Node node);
	
}
