/**
$Id: ZipFilePacker.java 8085 2022-10-27 05:25:30Z nicklas $

Copyright (C) 2007 Nicklas Nordborg, Martin Svensson

This file is part of BASE - BioArray Software Environment.
Available at http://base.thep.lu.se/

BASE is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License, or
(at your option) any later version.

BASE is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

package net.sf.basedb.util.zip;

import net.lingala.zip4j.io.outputstream.ZipOutputStream;
import net.lingala.zip4j.model.ZipParameters;
import net.lingala.zip4j.model.enums.AesKeyStrength;
import net.lingala.zip4j.model.enums.CompressionLevel;
import net.lingala.zip4j.model.enums.CompressionMethod;
import net.lingala.zip4j.model.enums.EncryptionMethod;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.StringParameterType;
import net.sf.basedb.util.FileUtil;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

/**
	This class packs files into a ZIP archive.

	@author Martin, Nicklas
	@version 2.4
	@base.modified $Date$
*/
public class ZipFilePacker 
	implements FilePacker
{
	private ZipOutputStream zip;
	private String password;
	
	/**
		Create a new packer that compresses files with the zip format.
	*/
	public ZipFilePacker()
	{}

	/*
		From the FilePacker interface
		-------------------------------------------
	*/
	/**
		@return Always "Zip-archive (.zip)"
	*/
	@Override
	public String getDescription()
	{
		return "Zip-archive (.zip)";
	}
	/**
	  @return Always "zip"
	*/
	@Override
	public String getFileExtension()
	{
		return "zip";
	}
	/**
	  	@return Always "application/zip"
	*/
	@Override
	public String getMimeType()
	{
		return "application/zip";
	}
	/**
		Wrap the output stream in a {@link ZipOutputStream}.
	*/
	@Override
	public void setOutputStream(OutputStream out)
	{
		try
		{
			this.zip = password == null ? new ZipOutputStream(out) : new ZipOutputStream(out, password.toCharArray());
		}
		catch (IOException ex)
		{
			throw new BaseException(ex);
		}
	}
	
	/**
		Encryption is supported.
		@since 3.19.5
	*/
	@Override
	public boolean supportsEncryption() 
	{
		return true;
	}

	/**
		Get the encyption password parameter.
		@since 3.19.5
	*/
	@Override
	public PluginParameter<String> getPasswordParameter()
	{
		PluginParameter<String> passwordParameter = new PluginParameter<String>
		(
			"password", "Password", "Enter a password to encrypt the archive with strong AES256 encryption. "
				+ "If empty, the ZIP archive will not be encrypted." , new StringParameterType()
		);
		return passwordParameter;
	}

	/**
		@since 3.19.5
	*/
	@Override
	public void setPassword(String password) 
	{
		this.password = password;
	}

	/**
		Create a new entry and write the compressed data to it.
	*/
	@Override
	public void pack(String entryName, InputStream in, long size, long lastModified)
		throws IOException
	{
		boolean isDirectory = in == null;
		if (isDirectory && !entryName.endsWith("/")) entryName += "/";
		
		ZipParameters zipParameters = new ZipParameters();
		zipParameters.setFileNameInZip(entryName);
		if (lastModified > 0)
		{
			zipParameters.setLastModifiedFileTime(lastModified);
		}
		if (isDirectory)
		{
			zipParameters.setCompressionMethod(CompressionMethod.STORE);
			zipParameters.setEntrySize(0);
			zipParameters.setEntryCRC(0);
		}
		else
		{
			zipParameters.setCompressionMethod(CompressionMethod.DEFLATE);
			zipParameters.setCompressionLevel(CompressionLevel.NORMAL);
			zipParameters.setEntrySize(size);
			if (password != null)
			{
				zipParameters.setEncryptFiles(true);
				zipParameters.setEncryptionMethod(EncryptionMethod.AES);
				zipParameters.setAesKeyStrength(AesKeyStrength.KEY_STRENGTH_256);
			}
		}
		zip.putNextEntry(zipParameters);
		if (!isDirectory) FileUtil.copy(in, zip);
		zip.flush();
		zip.closeEntry();
	}
	/**
		Finish the ZIP file.
	*/
	@Override
	public void close()
		throws IOException
	{
		zip.close();
		zip = null;
	}
	// -------------------------------------------

}
